"""
__author__ = "Jon Waggoner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Jon Waggoner"
__email__ = "opensource@jonwaggoner.com"
"""

import pathspec
from . import logger


def default_ignore_list():
    return [".DS_Store", "ascmhl", "ascmhl/"]


class MHLIgnoreSpec:
    """
    MHLIgnoreSpec wraps the parsing of ignore patterns to create a pathspec.PathSpec instance for filepath matching.
    """

    def __init__(self, existing_pattern_list=None, new_pattern_list=None, new_pattern_file=None):
        self._ignore_list = []
        self.set_patterns(existing_pattern_list, new_pattern_list, new_pattern_file)

    def set_patterns(self, existing_pattern_list=None, new_pattern_list=None, new_pattern_file=None):
        """
        set_patterns will set self.ignore_list to a new empty list then expand it with the supplied params.
        new patterns are applied on top of existing patterns.
        if no existing pattern, the default ignore pattern is used.
        the pattern file is applied on top of the pattern lists.
        no duplicates will be added.
        """
        self._ignore_list = []
        if existing_pattern_list:
            self._append_patterns_list(existing_pattern_list)
        else:
            self._append_patterns_list(default_ignore_list())
        if new_pattern_list:
            self._append_patterns_list(new_pattern_list)
        if new_pattern_file:
            self._append_patterns_from_file(new_pattern_file)

    def get_path_spec(self):
        """
        get_path_spec will return a pathspec.PathSpec instance filled with the contents of self.ignore_list
        the returned pathspec.PathSpec instance can be used to match against filepaths.
        """
        return pathspec.PathSpec.from_lines("gitwildmatch", iter(self._ignore_list))

    def get_pattern_list(self):
        return self._ignore_list.copy()

    def _append_patterns_list(self, patterns_to_append):
        """
        _append_patterns_list extends self.ignore_list with the contents of patterns_to_append.
        duplicates are ignored.
        """
        if patterns_to_append:
            self._ignore_list.extend(line for line in patterns_to_append if line not in self._ignore_list)

    def _append_patterns_from_file(self, filepath):
        """
        _append_patterns_from_file reads each line from a file then sends them to _append_patterns_list()
        """
        patters_from_file = []
        if filepath:
            with open(filepath, "r") as fh:
                patters_from_file.extend(line.rstrip("\n") for line in fh if line != "\n")
                self._append_patterns_list(patters_from_file)

    # For call to repr().
    def __repr__(self):
        return repr(self._ignore_list)

    # For call to str().
    def __str__(self):
        return str(self._ignore_list)



"""
__author__ = "Jon Waggoner, Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import pytest
from ascmhl.hasher import *


def test_cannot_instantiate_abstract_classes():
    with pytest.raises(TypeError):
        h = Hasher()  # this should raise TypeError - abstract class
    with pytest.raises(TypeError):
        h = HexHasher()  # this should raise TypeError - abstract class


def test_hash_data():
    # the data to hash
    data = b"media-hash-list"
    # map of hash algorithm to expected hash value
    hash_type_and_value = {
        "md5": "9db0fc9f30f5ee70041a7538809e2858",
        "sha1": "7b57673ac5633937a55b59009ad0c57ee08188b7",
        "xxh32": "f67c5a4f",
        "xxh64": "584b2ea1974f2b7c",
        "xxh3": "6d4cbd75905c81aa",
        "xxh128": "61a67c014f703a456ee7a776fd8c06bd",
        "c4": "c456LycWwpMMS7VDZEKvYv2L1uJS6s4qAFnaJdnQiy5JVbBFZMA8aLDS6SPaJjLqxXH4qZdnbuktopMt9frtC2qL1R",
    }
    # ensure all of our hash types produce the expected hash value for the data
    for k, v in hash_type_and_value.items():
        h = hash_data(data, k)
        assert h == v  # assert our computed hash equals expected hash


def test_hash_file(fs):
    # write some data to a file so we can hash it.
    file, data = "/data-file.txt", "media-hash-list"
    fs.create_file(file, contents=data)
    # map of hash algorithm to expected hash value
    hash_type_and_value = {
        "md5": "9db0fc9f30f5ee70041a7538809e2858",
        "sha1": "7b57673ac5633937a55b59009ad0c57ee08188b7",
        "xxh32": "f67c5a4f",
        "xxh64": "584b2ea1974f2b7c",
        "xxh3": "6d4cbd75905c81aa",
        "xxh128": "61a67c014f703a456ee7a776fd8c06bd",
        "c4": "c456LycWwpMMS7VDZEKvYv2L1uJS6s4qAFnaJdnQiy5JVbBFZMA8aLDS6SPaJjLqxXH4qZdnbuktopMt9frtC2qL1R",
    }
    # ensure all of our hash types produce the expected hash value for the file data
    for k, v in hash_type_and_value.items():
        h = hash_file(file, k)
        assert h == v  # assert our computed hash equals expected hash


def test_hash_of_hash_list():
    # the hash list to hash - purposefully in reverse order here to ensure data ends up sorted.
    hash_list = ["bb", "aa10", "aa01", "aa"]
    # map of hash algorithm to expected hash string
    hash_type_and_value = {
        "md5": "88abf5c156ff637c6380ae461fc50c3f",
        "sha1": "22ad6b98111f6571c9ae1449988260b920960c16",
        "xxh32": "48d7bd5a",
        "xxh64": "c4e803e454029ee4",
        "xxh3": "d908e88da5ad4928",
        "xxh128": "9041108a28887a9d35b4eac5bbaa5794",
        # TODO: add c4 in here - it doesn't like my hash list above due to c4 encoder requiring length
    }
    # ensure all of our hash types produce the expected hash value for the hash list
    for k, v in hash_type_and_value.items():
        h = hash_of_hash_list(hash_list, k)
        assert h == v  # assert our computed hash equals expected hash


def test_hash_list_concat_equals_chunking():
    # this is a test to prove that concatenating all strings then hashing is same as feeding them to hasher one by one
    # the hash list to hash
    hash_list = ["bb", "aa10", "aa01", "aa"]
    # ensure all of our hash types produce the expected hash value for the hash list
    for ht in HashType:
        if ht == HashType.c4:
            continue  # skipping c4 on this test for same reason as test_hash_of_hash_list
        hasher = new_hasher_for_hash_type(ht.name)
        h1 = hasher.hash_of_hash_list(hash_list)
        concat = hasher.bytes_from_string_digest("".join(hash_list))
        h2 = hasher.hash_data(concat)
        assert h1 == h2  # assert that sequentially feeding hashes is same as concat then feeding hashes



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import pytest
from freezegun import freeze_time
from click.testing import CliRunner
import ascmhl.commands
import os
import time
import platform

# this file is automatically loaded by pytest we setup various shared fixtures here


@pytest.fixture(scope="session", autouse=True)
def set_timezone():
    """Fakes the host timezone to UTC so we don't get different mhl files if the tests run on different time zones
    seems like freezegun can't handle timezones like we want"""
    if not os.name == "nt":
        os.environ["TZ"] = "UTZ"
        time.tzset()
    else:
        os.system('tzutil /s "UTC"')


@pytest.fixture(autouse=True)
def setup_environment(monkeypatch):
    def fake_hostname():
        return "myHost.local"

    monkeypatch.setattr(platform, "node", fake_hostname)
    # TODO: also patch ascmhl_tool_version ?


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def nested_mhl_histories(fs):

    # create mhl histories on different directly levels
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, ["/root", "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    fs.create_file("/root/A/AB/AB1.txt", contents="AB1\n")
    result = runner.invoke(ascmhl.commands.create, ["/root/A/AA", "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/B/B1.txt", contents="B1\n")
    result = runner.invoke(ascmhl.commands.create, ["/root/B", "-h", "xxh64"])
    assert result.exit_code == 0

    fs.create_file("/root/B/BA/BA1.txt", contents="BA1\n")
    fs.create_file("/root/B/BB/BB1.txt", contents="BB1\n")
    result = runner.invoke(ascmhl.commands.create, ["/root/B/BB", "-h", "xxh64"])
    assert result.exit_code == 0


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def simple_mhl_history(fs):

    # create a simple mhl history with two files in one generation
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, ["/root", "-h", "xxh64"])
    assert result.exit_code == 0


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def simple_mhl_folder(fs):

    # create a simple folder structure with two files
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from click.testing import CliRunner
from freezegun import freeze_time

import ascmhl.commands


@freeze_time("2020-01-16 09:15:00")
def test_simple(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.flatten, ["/root/", "/out/"])
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_add_one_file_same_hashformat(fs, simple_mhl_history):
    runner = CliRunner()

    # add a sidecar
    fs.create_file("/root/sidecar.txt", contents="sidecar\n")
    runner = CliRunner()
    runner.invoke(ascmhl.commands.create, ["/root", "-h", "xxh64"])

    result = runner.invoke(ascmhl.commands.flatten, ["/root/", "/out/"])
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_simple_two_hashformats(fs, simple_mhl_history):
    runner = CliRunner()

    # add a sidecar
    runner = CliRunner()
    runner.invoke(ascmhl.commands.create, ["/root", "-h", "md5"])

    result = runner.invoke(ascmhl.commands.flatten, ["/root/", "/out/"])
    assert result.exit_code == 0



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import datetime
import os
import platform

import click
from lxml import etree

from . import logger
from . import errors
from . import ignore
from . import utils
from .ignore import MHLIgnoreSpec
from .__version__ import (
    ascmhl_supported_hashformats,
    ascmhl_folder_name,
    ascmhl_tool_name,
    ascmhl_tool_version,
    ascmhl_default_hashformat,
)
from .generator import MHLGenerationCreationSession
from .hasher import hash_file, DirectoryHashContext
from .hashlist import MHLMediaHash, MHLCreatorInfo, MHLProcessInfo, MHLTool, MHLProcess, MHLAuthor
from .history import MHLHistory
from .traverse import post_order_lexicographic


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
# general options
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
@click.option(
    "--hash_format",
    "-h",
    type=click.Choice(ascmhl_supported_hashformats),
    multiple=False,
    default=ascmhl_default_hashformat,
    help="Algorithm",
)
@click.option(
    "--no_directory_hashes",
    "-n",
    default=False,
    is_flag=True,
    help="Skip creation of directory hashes, only reference directories without hash",
)
# creatorinfo values
@click.option(
    "--author_name",
    default=None,
    help="Name value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_email",
    default=None,
    help="Email value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_phone",
    default=None,
    help="Phone value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_role",
    default=None,
    help="Role value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--location",
    default=None,
    help="Value for the <location> element in the <creatorinfo> element",
)
@click.option(
    "--comment",
    default=None,
    help="Value for the <comment> element in the <creatorinfo> element",
)
# subcommands
@click.option(
    "--single_file",
    "-sf",
    multiple=True,
    type=click.Path(exists=True),
    help="Record single file, no completeness check (multiple occurrences possible for adding multiple files",
)
@click.option(
    "ignore_list",
    "--ignore",
    "-i",
    multiple=True,
    help="A single file pattern to ignore.",
)
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
def create(
    root_path,
    verbose,
    hash_format,
    no_directory_hashes,
    single_file,
    ignore_list,
    ignore_spec_file,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
):
    """
    Create a new generation for a folder or file(s)

    \b
    The create command hashes all files given and creates a new generation in the
    mhl-history with records for all hashed files. The command compares the hashes
    against the hashes stored in previous generations if available.
    """
    # distinguish different behavior for entire folder vs single files
    if single_file is not None and len(single_file) > 0:
        create_for_single_files_subcommand(
            root_path,
            verbose,
            hash_format,
            single_file,
            author_name,
            author_email,
            author_phone,
            author_role,
            location,
            comment,
            ignore_list,
            ignore_spec_file,
        )
        return
    create_for_folder_subcommand(
        root_path,
        verbose,
        hash_format,
        no_directory_hashes,
        author_name,
        author_email,
        author_phone,
        author_role,
        location,
        comment,
        ignore_list,
        ignore_spec_file,
    )
    return


def create_for_folder_subcommand(
    root_path,
    verbose,
    hash_format,
    no_directory_hashes,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
    ignore_list=None,
    ignore_spec_file=None,
):
    # command formerly known as "seal"
    """
    Creates a new generation with all files in a folder hierarchy.

    ROOT_PATH: the root path to use for the asc mhl history

    All files are hashed and will be compared to previous records in the `asc-mhl` folder if they exists.
    The command finds files that are registered in the `asc-mhl` folder but that are missing in the file system.
    Files that are existent in the file system but are not registered in the `asc-mhl` folder yet, are registered
    as new entries in the newly created generation(s).
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"Creating new generation for folder at path: {root_path} ...")

    existing_history = MHLHistory.load_from_path(root_path)

    # we collect all paths we expect to find first and remove every path that we actually found while
    # traversing the file system, so this set will at the end contain the file paths not found in the file system
    not_found_paths = existing_history.set_of_file_paths()

    # create the ignore specification
    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    # start a verification session on the existing history
    session = MHLGenerationCreationSession(existing_history, ignore_spec)

    num_failed_verifications = 0
    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
    dir_content_hash_mappings = {}
    dir_structure_hash_mappings = {}
    for folder_path, children in post_order_lexicographic(root_path, session.ignore_spec.get_path_spec()):
        # generate directory hashes
        dir_hash_context = None
        if not no_directory_hashes:
            dir_hash_context = DirectoryHashContext(hash_format)
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            not_found_paths.discard(file_path)
            if is_dir:
                if not dir_hash_context:
                    continue
                if dir_hash_context:
                    dir_hash_context.append_directory_hashes(
                        file_path, dir_content_hash_mappings.pop(file_path), dir_structure_hash_mappings.pop(file_path)
                    )
            else:
                hash_string, success = seal_file_path(existing_history, file_path, hash_format, session)
                if not success:
                    num_failed_verifications += 1
                if not no_directory_hashes:
                    dir_hash_context.append_file_hash(file_path, hash_string)
        dir_content_hash = None
        dir_structure_hash = None
        if dir_hash_context:
            dir_content_hash = dir_hash_context.final_content_hash_str()
            dir_structure_hash = dir_hash_context.final_structure_hash_str()
            dir_content_hash_mappings[folder_path] = dir_content_hash
            dir_structure_hash_mappings[folder_path] = dir_structure_hash
        modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(folder_path))
        session.append_directory_hashes(
            folder_path, modification_date, hash_format, dir_content_hash, dir_structure_hash
        )

    commit_session(session, author_name, author_email, author_phone, author_role, location, comment)

    exception = test_for_missing_files(not_found_paths, root_path, ignore_spec)
    if num_failed_verifications > 0:
        exception = errors.VerificationFailedException()

    if exception:
        raise exception


def create_for_single_files_subcommand(
    root_path,
    verbose,
    hash_format,
    single_file,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
    ignore_list=None,
    ignore_spec_file=None,
):
    # command formerly known as "record"
    """
    Creates a new generation with the given file(s) or folder(s).

    \b
    ROOT_PATH: the root path to use for the asc mhl history
    single_file: one or multiple paths to files or folders to be recorded

    All files that are specified or the files inside a specified folder are hashed and will be compared
    to previous records in the `asc-mhl` folder if they are recorded in the history already.

    The following files will not be handled by this command:

    \b
    * files that are referenced in the existing ascmhl history but not specified as input
    * files that are neither referenced in the history nor specified as input
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    assert len(single_file) != 0

    existing_history = MHLHistory.load_from_path(root_path)
    # start a creation session on the existing history
    session = MHLGenerationCreationSession(existing_history)

    num_failed_verifications = 0
    for path in single_file:
        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)
        if os.path.isdir(path):
            for folder_path, children in post_order_lexicographic(path, session.ignore_spec.get_path_spec()):
                for item_name, is_dir in children:
                    file_path = os.path.join(folder_path, item_name)
                    if is_dir:
                        continue
                    _, success = seal_file_path(existing_history, file_path, hash_format, session)
                    if not success:
                        num_failed_verifications += 1
        else:
            _, success = seal_file_path(existing_history, path, hash_format, session)
            if not success:
                num_failed_verifications += 1

    commit_session(session, author_name, author_email, author_phone, author_role, location, comment)

    if num_failed_verifications > 0:
        raise errors.VerificationFailedException()


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
@click.option(
    "ignore_list",
    "--ignore",
    "-i",
    multiple=True,
    help="A single file pattern to ignore.",
)
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
# subcommand
@click.option(
    "--directory_hash",
    "-dh",
    default=False,
    is_flag=True,
    help="Just create and verify directory hashes.",
)
# options for directory_hash
@click.option(
    "--calculate_only",
    "-co",
    default=False,
    is_flag=True,
    help="Only calculate and print hashes, don't compare against history.",
)
@click.option(
    "--root_only",
    "-ro",
    default=False,
    is_flag=True,
    help=(
        "Only calculate and print root directory hash, don't compare against history (only applies to"
        " --directory_hash)."
    ),
)
# options
@click.option(
    "--hash_format",
    "-h",
    type=click.Choice(ascmhl_supported_hashformats),
    multiple=False,
    help="Algorithm for directory hashes",
)
# subcommands
@click.option(
    "--single_file",
    "-sf",
    multiple=False,
    type=click.Path(),
    help="Verify a single file",
)
@click.option(
    "--packing_list", "-pl", default=None, type=click.Path(exists=True), help="Verify against an external packing list"
)
def verify(
    root_path,
    verbose,
    directory_hash,
    hash_format,
    single_file,
    packing_list,
    ignore_list,
    ignore_spec_file,
    calculate_only,
    root_only,
):
    """
    Verify a folder, single file(s), or a directory hash

    \b
    The verify command is used to check files in the file system with records in
    the ASC MHL history. All given files are hashed and hash values are compared
    against hash values stored in the ASC MHL history. Missing files or additional
    files in the file system are reported as errors. No new ASC MHL file /
    generation is created.
    """

    if packing_list is not None:
        verify_entire_folder(
            root_path, verbose, single_file, packing_list, ignore_list, ignore_spec_file, calculate_only
        )
        return

    if directory_hash is True:
        verify_directory_hash_subcommand(
            root_path, verbose, hash_format, ignore_list, ignore_spec_file, calculate_only, root_only
        )
        return

    verify_entire_folder(root_path, verbose, single_file, None, ignore_list, ignore_spec_file)
    return


def verify_entire_folder(
    root_path, verbose, single_file, packing_list_path, ignore_list=None, ignore_spec_file=None, calculate_only=None
):
    """
    Checks MHL hashes from all generations / a packing list against all file hashes.

    ROOT_PATH: the root path to use for the asc mhl history
    packing_list (opt): the path to the packing list

    Traverses through the content of a folder, hashes all found files and compares ("verifies") the hashes
    against the records in the asc-mhl folder/packing list. The command finds all files that are existent in
    the file system but are not registered in the `asc-mhl` folder/packing list yet, and all files that are
    registered in the `asc-mhl` folder/packing list but that are missing in the file system.
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    if single_file is not None and not os.path.isabs(single_file):
        single_file = os.path.join(root_path, single_file)

    logger.verbose(f"check folder at path: {root_path}")

    if packing_list_path is not None:
        existing_history = MHLHistory.load_from_packing_list_path(packing_list_path, root_path)
    else:
        existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    # we collect all paths we expect to find first and remove every path that we actually found while
    # traversing the file system, so this set will at the end contain the file paths not found in the file system
    not_found_paths = existing_history.set_of_file_paths()

    num_failed_verifications = 0
    num_new_files = 0

    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    found_single_file = False

    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            not_found_paths.discard(file_path)
            relative_path = existing_history.get_relative_file_path(file_path)
            history, history_relative_path = existing_history.find_history_for_path(relative_path)
            if is_dir:
                # TODO: find new directories here
                continue

            if single_file is None or os.path.realpath(single_file) == os.path.realpath(file_path):

                # check if there is an existing hash in the other generations and verify
                original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

                # in case there is no original hash entry continue
                if original_hash_entry is None:
                    logger.error(f"found new file {relative_path}")
                    num_new_files += 1
                    continue

                # create a new hash and compare it against the original hash entry
                current_hash = hash_file(file_path, original_hash_entry.hash_format)
                if original_hash_entry.hash_string == current_hash:
                    logger.verbose(f"verification ({original_hash_entry.hash_format}) of file {relative_path}: OK")
                else:
                    logger.error(
                        f"ERROR: hash mismatch        for {relative_path} "
                        f"old {original_hash_entry.hash_format}: {original_hash_entry.hash_string}, "
                        f"new {original_hash_entry.hash_format}: {current_hash}"
                    )
                    num_failed_verifications += 1

                found_single_file = True

    exception = test_for_missing_files(not_found_paths, root_path, ignore_spec)

    if found_single_file == False:
        exception = errors.SingelFileNotFoundException()

    if num_new_files > 0:
        exception = errors.NewFilesFoundException()
    if num_failed_verifications > 0:
        exception = errors.VerificationFailedException()

    if exception:
        raise exception


def verify_directory_hash_subcommand(
    root_path, verbose, hash_format, ignore_list=None, ignore_spec_file=None, calculate_only=False, root_only=False
):
    """
    Checks MHL directory hashes from all generations against computed directory hashes.

    ROOT_PATH: the root path to use for the asc mhl history

    Traverses through the content of a folder, hashes all found files, create directory hashes, and compares
    ("verifies") the hashes against the directory hash records in the asc-mhl folder.
    Content directory hashes and structure directory hashes are compared individually.
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"check folder at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    # choose the hash format of the latest root directory hash
    if hash_format is None:
        generation = -1
        for hash_list in existing_history.hash_lists:
            if hash_list.generation_number > generation:
                if len(hash_list.process_info.root_media_hash.hash_entries) > 0:
                    hash_format = hash_list.process_info.root_media_hash.hash_entries[0].hash_format

        if hash_format is None:
            logger.verbose(f"default hash format: c4")
            hash_format = "c4"
        else:
            logger.verbose(f"hash format from latest generation with directory hashes: {hash_format}")
    else:
        logger.verbose(f"hash format: {hash_format}")

    # start a verification session on the existing history
    session = MHLGenerationCreationSession(existing_history)

    num_failed_verifications = 0
    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
    dir_content_hash_mappings = {}
    dir_structure_hash_mappings = {}
    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        # generate directory hashes
        dir_hash_context = None
        dir_hash_context = DirectoryHashContext(hash_format)
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            if is_dir:
                file_path = os.path.join(folder_path, item_name)
                relative_path = existing_history.get_relative_file_path(file_path)
                history, history_relative_path = existing_history.find_history_for_path(relative_path)
                # check if there are directory hashes in the generations
                directory_hash_entries = history.find_directory_hash_entries_for_path(history_relative_path)

                content_hash = dir_content_hash_mappings.pop(file_path)
                structure_hash = dir_structure_hash_mappings.pop(file_path)
                dir_hash_context.append_directory_hashes(file_path, content_hash, structure_hash)

                num_successful_verifications = 0
                found_hash_format = False
                for directory_hash_entry in directory_hash_entries:
                    if directory_hash_entry.hash_format != hash_format:
                        continue
                    found_hash_format = True

                    num_current_successful_verifications = -1
                    if not root_only:
                        num_current_successful_verifications = _compare_and_log_directory_hashes(
                            relative_path, directory_hash_entry, content_hash, structure_hash
                        )

                    if num_current_successful_verifications == 2:
                        num_successful_verifications += 1
                    if num_current_successful_verifications == 1:
                        num_failed_verifications += 1

                if not calculate_only:
                    if not found_hash_format:
                        logger.error(
                            f"ERROR: verification of folder {relative_path}: No directory hash of type"
                            f" {hash_format} found"
                        )
                        num_failed_verifications += 1
            else:
                hash_string = hash_file(file_path, hash_format)
                dir_hash_context.append_file_hash(file_path, hash_string)
        dir_content_hash = None
        dir_structure_hash = None
        if dir_hash_context:
            dir_content_hash = dir_hash_context.final_content_hash_str()
            dir_structure_hash = dir_hash_context.final_structure_hash_str()
            dir_content_hash_mappings[folder_path] = dir_content_hash
            dir_structure_hash_mappings[folder_path] = dir_structure_hash
        modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(folder_path))

        logger.verbose_logging = calculate_only
        relative_path = session.root_history.get_relative_file_path(folder_path)
        if root_only and relative_path != ".":
            logger.verbose_logging = verbose
        session.append_directory_hashes(
            folder_path, modification_date, hash_format, dir_content_hash, dir_structure_hash
        )
        logger.verbose_logging = verbose

        # compare root hashes, works differently
        if folder_path == root_path:
            found_hash_format = False
            for hash_list in existing_history.hash_lists:
                root_hash_entries = hash_list.process_info.root_media_hash.hash_entries
                if len(root_hash_entries) > 0:
                    for root_hash_entry in root_hash_entries:
                        if root_hash_entry.hash_format == hash_format:
                            _compare_and_log_directory_hashes(
                                ".", root_hash_entry, dir_content_hash, dir_structure_hash
                            )
                            found_hash_format = True
            if not calculate_only:
                if not found_hash_format:
                    logger.error(f"ERROR: verification of root folder: No directory hash of type {hash_format} found")

    exception = None
    if num_failed_verifications > 0:
        exception = errors.VerificationDirectoriesFailedException()

    if exception:
        raise exception


def _compare_and_log_directory_hashes(
    relative_path, directory_hash_entry, calculated_content_hash_string, calculated_structure_hash_string
):
    num_successful_verifications = 0
    root_string = ""
    if hasattr(directory_hash_entry, "temp_is_root_folder") and directory_hash_entry.temp_is_root_folder:
        root_string = " (root folder in child history)"
    if (
        directory_hash_entry.hash_string == calculated_content_hash_string
        and directory_hash_entry.structure_hash_string == calculated_structure_hash_string
    ):
        if relative_path == ".":
            logger.verbose(
                f"  verification of root folder   OK (generation {directory_hash_entry.temp_generation_number:04d})"
            )
        else:
            logger.verbose(
                f"  verification of folder        {relative_path}{root_string} OK "
                f"(generation {directory_hash_entry.temp_generation_number:04d})"
            )

        num_successful_verifications += 2
    else:
        if directory_hash_entry.hash_string != calculated_content_hash_string:
            logger.error(
                f"ERROR: content hash mismatch   for {relative_path}{root_string} "
                f"old {directory_hash_entry.hash_format}: {directory_hash_entry.hash_string}, "
                f"new {directory_hash_entry.hash_format}: {calculated_content_hash_string} "
                f"(generation {directory_hash_entry.temp_generation_number:04d})"
            )
        else:
            logger.verbose(
                f"  content hash matches for      {relative_path}{root_string} "
                f" {directory_hash_entry.hash_format}: {directory_hash_entry.hash_string}"
                f" (generation {directory_hash_entry.temp_generation_number:04d})"
            )

        if directory_hash_entry.structure_hash_string != calculated_structure_hash_string:
            logger.error(
                f"ERROR: structure hash mismatch for {relative_path}{root_string} "
                f"old {directory_hash_entry.hash_format}: {directory_hash_entry.structure_hash_string}, "
                f"new {directory_hash_entry.hash_format}: {calculated_structure_hash_string} "
                f"(generation {directory_hash_entry.temp_generation_number:04d})"
            )
        else:
            logger.verbose(
                f"  structure hash matches for    {relative_path}{root_string} "
                f" {directory_hash_entry.hash_format}: {directory_hash_entry.hash_string} "
                f" (generation {directory_hash_entry.temp_generation_number:04d})"
            )

        num_successful_verifications += 1

    return num_successful_verifications


# TODO def verify_single_file_subcommand(root_path, verbose):


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
@click.option(
    "ignore_list",
    "--ignore",
    "-i",
    multiple=True,
    help="A single file pattern to ignore.",
)
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
def diff(root_path, verbose, ignore_list, ignore_spec_file):
    """
    Diff an entire folder structure

    \b
    The diff command is used to quickly compare files in the file system with
    records in the ASC MHL history. In comparison to the verify command, no
    hash values are created and compared. Missing files or additional files
    in the file system are reported as errors. No new ASC MHL file / generation
    is created.
    """
    diff_entire_folder_against_full_history_subcommand(root_path, verbose, ignore_list, ignore_spec_file)
    return


def diff_entire_folder_against_full_history_subcommand(root_path, verbose, ignore_list=None, ignore_spec_file=None):
    """
    Checks MHL hashes from all generations against all file hash entries.

    ROOT_PATH: the root path to use for the asc mhl history

    Traverses through the content of a folder. The command finds all files that are existent in the file system
    but are not registered in the `asc-mhl` folder yet, and all files that are registered in the `asc-mhl` folder
    but that are missing in the file system.
    """
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"check folder at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    # we collect all paths we expect to find first and remove every path that we actually found while
    # traversing the file system, so this set will at the end contain the file paths not found in the file system
    not_found_paths = existing_history.set_of_file_paths()
    num_failed_verifications = 0
    num_new_files = 0

    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
        for item_name, is_dir in children:
            file_path = os.path.join(folder_path, item_name)
            not_found_paths.discard(file_path)
            relative_path = existing_history.get_relative_file_path(file_path)
            history, history_relative_path = existing_history.find_history_for_path(relative_path)
            if is_dir:
                # TODO: find new directories here
                continue

            # check if there is an existing hash in the other generations and verify
            original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

            # in case there is no original hash entry continue
            if original_hash_entry is None:
                logger.error(f"found new file {relative_path}")
                num_new_files += 1
                continue

    exception = test_for_missing_files(not_found_paths, root_path, ignore_spec)
    if num_new_files > 0:
        exception = errors.NewFilesFoundException()
    if num_failed_verifications > 0:
        exception = errors.VerificationFailedException()

    if exception:
        raise exception


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.argument("destination_path", type=click.Path())
# general options
@click.option("--verbose", "-v", default=False, is_flag=True, help="Verbose output")
@click.option(
    "--no_directory_hashes",
    "-n",
    default=False,
    is_flag=True,
    help="Skip creation of directory hashes, only reference directories without hash",
)
@click.option("ignore_list", "--ignore", "-i", multiple=True, help="A single file pattern to ignore.")
@click.option(
    "ignore_spec_file",
    "--ignore_spec",
    "-ii",
    type=click.Path(exists=True),
    help="A file containing multiple file patterns to ignore.",
)
# creatorinfo values
@click.option(
    "--author_name",
    default=None,
    help="Name value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_email",
    default=None,
    help="Email value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_phone",
    default=None,
    help="Phone value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--author_role",
    default=None,
    help="Role value for the <author> element in the <creatorinfo> element",
)
@click.option(
    "--location",
    default=None,
    help="Value for the <location> element in the <creatorinfo> element",
)
@click.option(
    "--comment",
    default=None,
    help="Value for the <comment> element in the <creatorinfo> element",
)
def flatten(
    root_path,
    destination_path,
    verbose,
    no_directory_hashes,
    ignore_list,
    ignore_spec_file,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
):
    """
    Flatten an MHL history into one external manifest

    \b
    The flatten command iterates through the mhl-history, collects all known files and
    their hashes in multiple hash formats and writes them to a new mhl file outside of the
    iterated history.
    """
    flatten_history(
        root_path,
        destination_path,
        verbose,
        no_directory_hashes,
        author_name,
        author_email,
        author_phone,
        author_role,
        location,
        comment,
        ignore_list,
        ignore_spec_file,
    )
    return


def flatten_history(
    root_path,
    destination_path,
    verbose,
    no_directory_hashes,
    author_name,
    author_email,
    author_phone,
    author_role,
    location,
    comment,
    ignore_list=None,
    ignore_spec_file=None,
):
    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.verbose(f"Flattening folder at path: {root_path} ...")

    existing_history = MHLHistory.load_from_path(root_path)

    # create the ignore specification
    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)

    # start a verification session on the existing history
    collection_history = MHLHistory.create_collection_at_path(destination_path)
    session = MHLGenerationCreationSession(collection_history, ignore_spec)

    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
    dir_hash_mappings = {}

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    for hash_list in existing_history.hash_lists:
        for media_hash in hash_list.media_hashes:
            if not media_hash.is_directory:
                for hash_entry in media_hash.hash_entries:
                    if hash_entry.action != "failed":
                        # check if this entry is newer than the one already in there to avoid duplicate entries
                        found_media_hash = session.new_hash_lists[collection_history].find_media_hash_for_path(
                            media_hash.path
                        )
                        if found_media_hash == None:
                            session.append_file_hash(
                                media_hash.path,
                                media_hash.file_size,
                                media_hash.last_modification_date,
                                hash_entry.hash_format,
                                hash_entry.hash_string,
                                action=hash_entry.action,
                                hash_date=hash_entry.hash_date,
                            )
                        else:
                            hashformat_is_already_there = False
                            for found_hash_entry in found_media_hash.hash_entries:
                                if found_hash_entry.hash_format == hash_entry.hash_format:
                                    hashformat_is_already_there = True
                            if not hashformat_is_already_there:
                                # assuming that hash_entry of same type also has same hash_value ..
                                session.append_file_hash(
                                    media_hash.path,
                                    media_hash.file_size,
                                    media_hash.last_modification_date,
                                    hash_entry.hash_format,
                                    hash_entry.hash_string,
                                    action=hash_entry.action,
                                    hash_date=hash_entry.hash_date,
                                )

    commit_session_for_collection(
        session, root_path, author_name, author_email, author_phone, author_role, location, comment
    )


@click.command()
@click.argument("root_path", required=False, type=click.Path(exists=True))
# options
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Verbose output",
)
# subcommands
@click.option(
    "--single_file",
    "-sf",
    multiple=True,
    type=click.Path(exists=True),
    help="Info for single file",
)
def info(verbose, single_file, root_path):
    """
    Prints information from the ASC MHL history

    \b
    """
    if single_file is not None and len(single_file) > 0:
        if root_path == None:
            current_dir = os.path.dirname(os.path.abspath(single_file[0]))
            while current_dir != "/" and current_dir != "":
                asc_mhl_folder_path = os.path.join(current_dir, ascmhl_folder_name)
                if os.path.exists(asc_mhl_folder_path):
                    root_path = current_dir
                    break
                current_dir = os.path.dirname(current_dir)
        if root_path == None:
            raise errors.NoMHLHistoryExceptionForPath(single_file[0])
        else:
            info_for_single_file(root_path, verbose, single_file)
        return
    else:
        info_for_entire_history(root_path, verbose)
    return


def info_for_entire_history(root_path, verbose):
    """
    ROOT_PATH: the root path to use for the asc mhl history
    """

    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.info(f"Info with history at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    for hash_list in existing_history.hash_lists:
        if logger.verbose_logging == True:
            creatorInfo = hash_list.creator_info.summary()
            processInfo = hash_list.process_info.summary()
            logger.info(
                f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})\n"
                f"     CreatorInfo: {creatorInfo}\n"
                f"     ProcessInfo: {processInfo}"
            )
        else:
            logger.info(f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})")

        if len(existing_history.child_histories) > 0:
            if logger.verbose_logging == True:
                for child_history in existing_history.child_histories:
                    logger.info(f"\nChild History at {child_history.get_root_path()}:")
                    for hash_list in child_history.hash_lists:
                        if logger.verbose_logging == True:
                            creatorInfo = hash_list.creator_info.summary()
                            processInfo = hash_list.process_info.summary()
                            logger.info(
                                f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})\n"
                                f"     CreatorInfo: {creatorInfo}\n"
                                f"     ProcessInfo: {processInfo}"
                            )
                        else:
                            logger.info(
                                f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})"
                            )
            else:
                logger.info(f"  Child Histories: {len(existing_history.child_histories)}")


def info_for_single_file(root_path, verbose, single_file):
    """
    ROOT_PATH: the root path to use for the asc mhl history (optional)
    """

    logger.verbose_logging = verbose

    if not os.path.isabs(root_path):
        root_path = os.path.join(os.getcwd(), root_path)

    logger.info(f"Info with history at path: {root_path}")

    existing_history = MHLHistory.load_from_path(root_path)

    if len(existing_history.hash_lists) == 0:
        raise errors.NoMHLHistoryException(root_path)

    for path in single_file:
        relative_path = existing_history.get_relative_file_path(os.path.abspath(path))
        logger.info(f"{relative_path}:")
        for hash_list in existing_history.hash_lists:
            media_hash = hash_list.find_media_hash_for_path(relative_path)
            if media_hash is None:
                continue
            for hash_entry in media_hash.hash_entries:
                if logger.verbose_logging == True:
                    absolutePath = os.path.join(hash_list.get_root_path(), media_hash.path)
                    creatorInfo = hash_list.creator_info.summary()
                    processInfo = hash_list.process_info.summary()
                    logger.info(
                        f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})"
                        f" {hash_entry.hash_format}: {hash_entry.hash_string} ({hash_entry.action}) \n"
                        f"    {absolutePath}\n"
                        f"     CreatorInfo: {creatorInfo}\n"
                        f"     ProcessInfo: {processInfo}"
                    )
                else:
                    logger.info(
                        f"  Generation {hash_list.generation_number} ({hash_list.creator_info.creation_date})"
                        f" {hash_entry.hash_format}: {hash_entry.hash_string} ({hash_entry.action})"
                    )


@click.command()
@click.argument("file_path", type=click.Path(exists=True))
# subcommands
@click.option(
    "--directory_file",
    "-df",
    default=False,
    is_flag=True,
    help="Check directory file (e.g. ascmhl_chain.xml) instead of manifest file",
)
def xsd_schema_check(file_path, directory_file):
    """
    Checks a .mhl file against the xsd schema definition

    \b
    The xsd-schema-check command validates a given ASC MHL file against the XML
    XSD. This command can be used to ensure the creation of syntactically valid
    ASC MHL files, for example during implementation of tools creating ASC MHL
    files.
    """

    xsd_path = "xsd/ASCMHL.xsd"

    if directory_file:
        xsd_path = "xsd/ASCMHLDirectory__combined.xsd"

    xsd = etree.XMLSchema(etree.parse(xsd_path))

    # pass a file handle to support the fake file system used in the tests
    file = open(file_path, "rb")
    result = xsd.validate(etree.parse(file))

    if result:
        logger.info(f"validated: {file_path}")
    else:
        logger.error(f"ERROR: {file_path} didn't validate against XSD!")
        logger.info(f"Issues:\n{xsd.error_log}")
        raise errors.VerificationFailedException


def test_for_missing_files(not_found_paths, root_path, ignore_spec: MHLIgnoreSpec = MHLIgnoreSpec()):
    ignore_path_spec = ignore_spec.get_path_spec()
    # update to exclude our ignored files
    not_found_paths = [x for x in not_found_paths if not ignore_path_spec.match_file(x)]
    if len(not_found_paths) == 0:
        return None
    # test our not_found_paths against our ignore spec to ensure these weren't explicitly ignored.
    logger.error(f"ERROR: {len(not_found_paths)} missing file(s):")
    for path in not_found_paths:
        logger.error(f"  {os.path.relpath(path, root_path)}")
    return errors.CompletenessCheckFailedException()


def commit_session(session, author_name, author_email, author_phone, author_role, location, comment):
    creator_info = MHLCreatorInfo()
    creator_info.tool = MHLTool(ascmhl_tool_name, ascmhl_tool_version)
    creator_info.creation_date = utils.datetime_now_isostring()
    creator_info.host_name = platform.node()
    creator_info.location = location
    creator_info.comment = comment
    if author_name is not None:
        author_object = MHLAuthor(author_name, author_email, author_phone, author_role)
        creator_info.authors.append(author_object)

    process_info = MHLProcessInfo()
    process_info.process = MHLProcess("in-place")

    session.commit(creator_info, process_info)


def commit_session_for_collection(
    session, root_path, author_name, author_email, author_phone, author_role, location, comment
):
    creator_info = MHLCreatorInfo()
    creator_info.tool = MHLTool(ascmhl_tool_name, ascmhl_tool_version)
    creator_info.creation_date = utils.datetime_now_isostring()
    creator_info.host_name = platform.node()
    creator_info.location = location
    creator_info.comment = comment
    if author_name is not None:
        author_object = MHLAuthor(author_name, author_email, author_phone, author_role)
        creator_info.authors.append(author_object)

    process_info = MHLProcessInfo()
    process_info.process = MHLProcess("flatten")
    root_hash = MHLMediaHash()
    root_hash.path = root_path
    process_info.root_media_hash = root_hash
    process_info.hashlist_custom_basename = "packinglist_" + os.path.basename(root_path)

    session.commit(creator_info, process_info)


def seal_file_path(existing_history, file_path, hash_format, session) -> (str, bool):
    relative_path = existing_history.get_relative_file_path(file_path)
    file_size = os.path.getsize(file_path)
    file_modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(file_path))

    # find in the according child history the already available hash formats
    existing_child_history, existing_history_relative_path = existing_history.find_history_for_path(relative_path)
    existing_hash_formats = existing_child_history.find_existing_hash_formats_for_path(existing_history_relative_path)

    # in case there is no hash in the required format to use yet, we need to verify also against
    # one of the existing hash formats, we for simplicity use always the first hash format in this example
    # but one could also use a different one if desired
    success = True
    if len(existing_hash_formats) > 0 and hash_format not in existing_hash_formats:
        existing_hash_format = existing_hash_formats[0]
        hash_in_existing_format = hash_file(file_path, existing_hash_format)
        # FIXME: test what happens if the existing hash verification fails in other format fails
        # should we then really create two entries
        success &= session.append_file_hash(
            file_path, file_size, file_modification_date, existing_hash_format, hash_in_existing_format
        )
    current_format_hash = hash_file(file_path, hash_format)
    # in case the existing hash verification failed we don't want to add the current format hash to the generation
    # but we need to return it for directory hash creation
    if not success:
        return current_format_hash, False
    success &= session.append_file_hash(file_path, file_size, file_modification_date, hash_format, current_format_hash)
    return current_format_hash, success



"""
__author__ = "Jon Waggoner, Alexander Sahm, Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner"
__email__ = "opensource@pomfort.com"
"""
import binascii
import hashlib

import xxhash
import os
from enum import Enum, unique
from abc import ABC, abstractmethod


class Hasher(ABC):
    """
    Hasher is an abstract base class (ABC) that outlines the needed hash functionality by ascmhl.
    This abstraction is primarily necessary due to some discrepancies in the hash encoding of C4ID.
    """

    def __init__(self):
        # instantiate our internal hash generator. such as: hashlib.md5 or xxhash.xxh64
        self.hasher = self.hashlib_type()()

    def update(self, data: bytes) -> None:
        """
        writes data to the internal hasher.
        """
        self.hasher.update(data)

    @abstractmethod
    def string_digest(self) -> str:
        """
        get the string digest of the current state of the internal hasher
        """
        pass

    @classmethod
    @abstractmethod
    def bytes_from_string_digest(cls, hash_string: str) -> bytes:
        """
        helper to convert a hash string to its underlying byte representation adhering to the encoding of the hash_type.
        """
        pass

    @staticmethod
    @abstractmethod
    def hashlib_type():
        """
        returns the underlying wrapped hasher type from either the hashlib or xxhash libraries.
        such as: hashlib.md5 or xxhash.xxh64
        """
        pass

    @classmethod
    def hash_of_hash_list(cls, hash_list: [str]) -> str:
        """
        generates and returns a new hash string by hashing a list of existing hash strings.
        """
        hasher = cls()

        # empty checksum for empty lists
        if len(hash_list) == 0:
            return hasher.string_digest()

        # sort lexicographically
        hash_list.sort()
        for hash_string in hash_list:
            hasher.update(cls.bytes_from_string_digest(hash_string))

        return hasher.string_digest()

    @classmethod
    def hash_file(cls, filepath: str) -> str:
        """
        computes and returns a new hash string for a file

        arguments:
        filepath -- string value, path of file to generate hash for.
        hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
        """
        hasher = cls()
        with open(filepath, "rb") as fd:
            # process files in chunks so that large files won't cause excessive memory consumption.
            size = 1024 * 1024  # chunk size 1MB
            chunk = fd.read(size)
            while chunk:
                hasher.update(chunk)
                chunk = fd.read(size)

        return hasher.string_digest()

    @classmethod
    def hash_data(cls, input_data: bytes) -> str:
        """
        computes and returns a new hash string from the input data.

        arguments:
        input_data -- the bytes to compute the hash from
        hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
        """
        hasher = cls()
        hasher.update(input_data)
        return hasher.string_digest()


class HexHasher(Hasher, ABC):
    """
    HexHasher inherits from the base Hasher class to implement the hexadecimal encoding and decoding of hashes.
    Cannot be instantiated directly. Choose a subclass.
    """

    def string_digest(self) -> str:
        return self.hasher.hexdigest()

    @classmethod
    def bytes_from_string_digest(cls, hash_string: str) -> bytes:
        return binascii.unhexlify(hash_string)


class MD5(HexHasher):
    """
    md5 checksum generator.
    """

    @staticmethod
    def hashlib_type():
        return hashlib.md5


class SHA1(HexHasher):
    """
    sha1 checksum generator.
    """

    @staticmethod
    def hashlib_type():
        return hashlib.sha1


class XXH32(HexHasher):
    """
    xxh32 checksum generator.
    """

    @staticmethod
    def hashlib_type():
        return xxhash.xxh32


class XXH64(HexHasher):
    """
    xxh64 checksum generator.
    """

    @staticmethod
    def hashlib_type():
        return xxhash.xxh64


class XXH3(HexHasher):
    """
    xxh3 checksum generator.
    """

    @staticmethod
    def hashlib_type():
        return xxhash.xxh3_64


class XXH128(HexHasher):
    """
    xxh128 checksum generator.
    """

    @staticmethod
    def hashlib_type():
        return xxhash.xxh3_128


class C4(Hasher):
    """
    C4 checksum generator.
    C4 Hasher is different than the other supported hash algorithms in that it does not adhere to a hex char set.
    """

    # c4 has a different character set than the usual hex char set of other checksum types. encoding is different.
    charset = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"  # C4ID character set

    @staticmethod
    def hashlib_type():
        return hashlib.sha512

    def string_digest(self) -> str:
        sha512_string = self.hasher.hexdigest()

        base58 = 58  # the encoding basis
        c4id_length = 90  # the guaranteed length
        zero = "1"  # '0' is not in the C4ID alphabet so '1' is zero

        hash_value = int(sha512_string, 16)
        c4_string = ""
        while hash_value != 0:
            modulo = hash_value % base58
            hash_value = hash_value // base58
            c4_string = C4.charset[modulo] + c4_string

        c4_string = "c4" + c4_string.rjust(c4id_length - 2, zero)
        return c4_string

    @classmethod
    def bytes_from_string_digest(cls, hash_string: str) -> bytes:
        base58 = 58  # the encoding basis
        c4id_length = 90  # the guaranteed length
        result = 0
        i = 2

        while i < c4id_length:
            temp = C4.charset.index(hash_string[i])
            result = result * base58 + temp
            i = i + 1

        data = result.to_bytes(64, byteorder="big")
        return data


@unique
class HashType(Enum):
    """
    HashType wraps all ascmhl supported hash formats.
    """

    md5 = MD5
    sha1 = SHA1
    xxh32 = XXH32
    xxh64 = XXH64
    xxh3 = XXH3
    xxh128 = XXH128
    c4 = C4


class DirectoryHashContext:
    """
    DirectoryHashContext wraps the data necessary to compute directory checksums.
    """

    def __init__(self, hash_format: str):
        self.hash_format = hash_format
        self.hasher = new_hasher_for_hash_type(hash_format)
        self.content_hash_strings = []
        self.structure_hash_strings = []

    def append_file_hash(self, path: str, content_hash_string: str):
        """
        append child file data to this directory context.
        """
        self.content_hash_strings.append(content_hash_string)

        # structure hashes are computed from lists of children name+hash strings
        path_bytes = os.path.basename(os.path.normpath(path)).encode("utf8")
        hash_bytes = self.hasher.bytes_from_string_digest(content_hash_string)
        structure_hash = self.hasher.hash_data(path_bytes + hash_bytes)
        self.structure_hash_strings.append(structure_hash)

    def append_directory_hashes(self, path: str, content_hash_string: str, structure_hash_string: str):
        """
        append child directory data to this directory context.
        """
        self.content_hash_strings.append(content_hash_string)

        # structure hashes are computed from lists of children name+hash strings
        path_bytes = os.path.basename(os.path.normpath(path)).encode("utf8")
        hash_bytes = self.hasher.bytes_from_string_digest(structure_hash_string)
        structure_hash = self.hasher.hash_data(path_bytes + hash_bytes)
        self.structure_hash_strings.append(structure_hash)

    def final_content_hash_str(self):
        """
        compute and return the content hash of this directory context by hashing the child content hash list.
        """
        return self.hasher.hash_of_hash_list(self.content_hash_strings)

    def final_structure_hash_str(self):
        """
        compute and return the structure hash of this directory context by hashing the child structure hash list.
        """
        return self.hasher.hash_of_hash_list(self.structure_hash_strings)


def new_hasher_for_hash_type(hash_format: str) -> Hasher:
    """
    creates a new instance of the appropriate Hasher class based on the hash_format argument

    arguments:
    hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
    """
    if not hash_format:
        raise ValueError
    hash_type = HashType[hash_format]
    if not hash_type:
        raise ValueError

    return hash_type.value()  # instantiate and return a new Hasher of the specified HashType


def hash_of_hash_list(hash_list: [str], hash_format: str) -> str:
    """
    computes and returns a new hash string from a given list of hash strings.

    arguments:
    hash_list -- list of string hashes to compute a new single hash from
    hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
    """
    hasher = new_hasher_for_hash_type(hash_format)
    return hasher.hash_of_hash_list(hash_list)


def hash_file(filepath: str, hash_format: str) -> str:
    """
    computes and returns a new hash string for a file

    arguments:
    filepath -- string value, path of file to generate hash for.
    hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
    """
    hasher = new_hasher_for_hash_type(hash_format)
    return hasher.hash_file(filepath)


def hash_data(input_data: bytes, hash_format: str) -> str:
    """
    computes and returns a new hash string from the input data.

    arguments:
    input_data -- the bytes to compute the hash from
    hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
    """
    hasher = new_hasher_for_hash_type(hash_format)
    return hasher.hash_data(input_data)


def bytes_for_hash_string(hash_string: str, hash_format: str) -> bytes:
    """
    wraps the different Hasher string to byte conversions

    arguments:
    hash_string -- string value, the hash string to convert to bytes
    hash_format -- string value, one of the supported hash formats, e.g. 'md5', 'xxh64'
    """
    hasher = new_hasher_for_hash_type(hash_format)
    return hasher.bytes_from_string_digest(hash_string)



from packaging import version

from ascmhl.utils import datetime_now_filename_string
import datetime
import os, time


def test_time():
    datetime_now_filename = datetime_now_filename_string()
    datetime_now_filename_reference = datetime.datetime.strftime(datetime.datetime.utcnow(), "%Y-%m-%d_%H%M%S")

    print("datetime_now_filename:           " + datetime_now_filename)
    print("datetime_now_filename_reference: " + datetime_now_filename_reference)

    # seems always to be working, as tests run in UTC ?
    assert datetime_now_filename == datetime_now_filename_reference



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from click.testing import CliRunner
from freezegun import freeze_time

import ascmhl.commands


@freeze_time("2020-01-16 09:15:00")
def test_simple(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["-dh", "-co", "/root/"])
    assert (
        "calculated directory hash for A  xxh64: d3904ee76bba3d2a (content), 3fbe33b2924e26aa (structure)"
        in result.output
    )
    assert "calculated root hash  xxh64: ca56d22f064fdf1b (content), 2ccca3899111eabb (structure)" in result.output
    assert result.exit_code == 0

    result = runner.invoke(ascmhl.commands.verify, ["-dh", "-co", "-ro", "/root/"])
    assert "calculated directory hash" not in result.output
    assert "calculated root hash  xxh64: ca56d22f064fdf1b (content), 2ccca3899111eabb (structure)" in result.output
    assert result.exit_code == 0



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
import re
from freezegun import freeze_time
from click.testing import CliRunner

from ascmhl import hashlist_xml_parser
from ascmhl.__version__ import ascmhl_file_extension
from ascmhl.history import MHLHistory
import ascmhl.commands


def test_simple_parsing():
    path = "examples/scenarios/Output/scenario_01/travel_01/A002R2EC/ascmhl/0001_A002R2EC_2020-01-16_091500.mhl"
    hash_list = hashlist_xml_parser.parse(path)
    assert len(hash_list.media_hashes) > 0


@freeze_time("2020-01-16 09:15:00")
def test_child_history_parsing(fs, nested_mhl_histories):
    """ """

    root_history = MHLHistory.load_from_path("/root")
    assert len(root_history.child_histories) == 2

    aa_history = root_history.child_histories[0]
    b_history = root_history.child_histories[1]
    bb_history = root_history.child_histories[1].child_histories[0]

    assert aa_history.asc_mhl_path == "/root/A/AA/ascmhl"
    assert aa_history.parent_history == root_history
    assert b_history.asc_mhl_path == "/root/B/ascmhl"
    assert b_history.parent_history == root_history

    assert len(b_history.child_histories) == 1
    assert bb_history.asc_mhl_path == "/root/B/BB/ascmhl"
    assert bb_history.parent_history == b_history

    # check sub children mappings that map all transitive children and their relative path
    assert root_history.child_history_mappings["A/AA"] == aa_history
    assert root_history.child_history_mappings["B"] == b_history
    assert root_history.child_history_mappings["B/BB"] == bb_history
    assert b_history.child_history_mappings["BB"] == bb_history

    # check if the correct (child) histories are returned for a given path
    assert root_history.find_history_for_path("Stuff.txt")[0] == root_history
    assert root_history.find_history_for_path("A/AA/AA1.txt")[0] == aa_history
    assert root_history.find_history_for_path("A/AB/AB1.txt")[0] == root_history
    assert root_history.find_history_for_path("B/B1.txt")[0] == b_history
    assert root_history.find_history_for_path("B/BA/BA1.txt")[0] == b_history
    assert root_history.find_history_for_path("B/BB/BB1.txt")[0] == bb_history

    # the history object should only return the media hashes and hash entries it contains directly
    # if we need th entries from child histories we have to ask them directly
    assert root_history.find_original_hash_entry_for_path("Stuff.txt") is not None
    assert root_history.find_original_hash_entry_for_path("A/AA/AA1.txt") is None
    assert aa_history.find_original_hash_entry_for_path("AA1.txt") is not None


@freeze_time("2020-01-16 09:15:00")
def test_child_history_verify(fs, nested_mhl_histories):
    """ """

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, ["/root"], catch_exceptions=False)
    assert result.exit_code == 0

    assert os.path.isfile("/root/ascmhl/0002_root_2020-01-16_091500.mhl")
    assert os.path.isfile("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500.mhl")
    assert os.path.isfile("/root/B/ascmhl/0002_B_2020-01-16_091500.mhl")
    assert os.path.isfile("/root/B/BB/ascmhl/0002_BB_2020-01-16_091500.mhl")

    root_history = MHLHistory.load_from_path("/root")
    assert len(root_history.hash_lists) == 2

    assert root_history.hash_lists[1].media_hashes[1].path == "A/AB/AB1.txt"
    assert root_history.hash_lists[1].media_hashes[1].hash_entries[0].action == "original"
    assert root_history.hash_lists[1].media_hashes[5].path == "Stuff.txt"
    assert root_history.hash_lists[1].media_hashes[5].hash_entries[0].action == "verified"

    aa_history = root_history.child_histories[0]
    b_history = root_history.child_histories[1]
    bb_history = root_history.child_histories[1].child_histories[0]
    root_hash_list = root_history.hash_lists[-1]
    aa_hash_list = aa_history.hash_lists[-1]
    b_hash_list = b_history.hash_lists[-1]
    bb_hash_list = bb_history.hash_lists[-1]

    assert aa_history.latest_generation_number() == 2
    assert b_hash_list.media_hashes[0].path == "BA/BA1.txt"
    assert b_hash_list.media_hashes[3].path == "B1.txt"
    assert b_hash_list.media_hashes[0].hash_entries[0].action == "original"
    assert b_hash_list.media_hashes[3].hash_entries[0].action == "verified"

    # check that the mhl references are correct
    assert root_history.hash_lists[1].referenced_hash_lists[0] == aa_hash_list
    assert root_history.hash_lists[1].referenced_hash_lists[1] == b_hash_list
    assert b_hash_list.referenced_hash_lists[0] == bb_hash_list
    assert len(aa_hash_list.referenced_hash_lists) == 0

    # the media hashes of the directories that contain a history themselves should be both in the child history
    # as root media hash and in the parent history to represent the directory that contains the child history
    aa_dir_hash = root_hash_list.find_media_hash_for_path("A/AA").hash_entries[0].hash_string
    assert aa_dir_hash
    assert aa_hash_list.process_info.root_media_hash.hash_entries[0].hash_string == aa_dir_hash
    # the dir hash of BB is in the history of B not in the root history
    assert root_hash_list.find_media_hash_for_path("B/BB") is None
    bb_dir_hash = b_hash_list.find_media_hash_for_path("BB").hash_entries[0].hash_string
    assert bb_hash_list.process_info.root_media_hash.hash_entries[0].hash_string == bb_dir_hash
    # but the dir hash of B is also in the root history
    assert root_hash_list.find_media_hash_for_path("B")


@freeze_time("2020-01-16 09:15:00")
def test_child_history_partial_verification_ba_1_file(fs, nested_mhl_histories):
    """ """

    # create an additional file the record command will not add since we only pass it B1 as single file
    fs.create_file("/root/B/B2.txt", contents="B2\n")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, ["/root", "-sf", "/root/B/B1.txt"], catch_exceptions=False)
    assert result.exit_code == 0

    # two new generations have been written
    assert os.path.isfile("/root/ascmhl/0002_root_2020-01-16_091500.mhl")
    assert os.path.isfile("/root/B/ascmhl/0002_B_2020-01-16_091500.mhl")

    root_history = MHLHistory.load_from_path("/root")
    assert len(root_history.hash_lists) == 2

    aa_history = root_history.child_histories[0]
    b_history = root_history.child_histories[1]
    bb_history = root_history.child_histories[1].child_histories[0]

    # the root hash list only contains a mhl reference to the hash list of the B history, no media hashes
    assert len(root_history.hash_lists[1].media_hashes) == 0
    assert root_history.hash_lists[1].referenced_hash_lists[0] == b_history.hash_lists[1]
    # the B hash list contains the media hash of the verified file
    assert b_history.hash_lists[1].media_hashes[0].path == "B1.txt"
    assert b_history.hash_lists[1].media_hashes[0].hash_entries[0].action == "verified"
    # the created B2 file is not referenced in the B history, only B1
    assert len(b_history.hash_lists[1].media_hashes) == 1

    # the other histories don't have a new generation
    assert not os.path.isfile("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500.mhl")
    assert not os.path.isfile("/root/B/BB/ascmhl/0002_BB_2020-01-16_091500.mhl")
    assert aa_history.latest_generation_number() == 1
    assert bb_history.latest_generation_number() == 1


@freeze_time("2020-01-16 09:15:00")
def test_child_history_partial_verification_bb_folder(fs, nested_mhl_histories):
    """ """

    # create an additional file the record command will find because we pass it a folder
    fs.create_file("/root/B/BB/BB2.txt", contents="BB2\n")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, ["/root", "-sf", "/root/B/BB"])
    assert result.exit_code == 0

    assert os.path.isfile("/root/ascmhl/0002_root_2020-01-16_091500.mhl")
    assert os.path.isfile("/root/B/ascmhl/0002_B_2020-01-16_091500.mhl")
    assert os.path.isfile("/root/B/BB/ascmhl/0002_BB_2020-01-16_091500.mhl")

    root_history = MHLHistory.load_from_path("/root")
    assert len(root_history.hash_lists) == 2

    aa_history = root_history.child_histories[0]
    b_history = root_history.child_histories[1]
    bb_history = root_history.child_histories[1].child_histories[0]

    # the root and the B hash lists only contains a mhl reference to the hash list
    # down the folder hierarchy, no media hashes
    assert len(root_history.hash_lists[1].media_hashes) == 0
    assert root_history.hash_lists[1].referenced_hash_lists[0] == b_history.hash_lists[1]
    assert len(b_history.hash_lists[1].media_hashes) == 0
    assert b_history.hash_lists[1].referenced_hash_lists[0] == bb_history.hash_lists[1]

    # the BB hash list contains the media hash of the verified files in BB
    assert bb_history.hash_lists[1].media_hashes[0].path == "BB1.txt"
    assert bb_history.hash_lists[1].media_hashes[0].hash_entries[0].action == "verified"
    assert bb_history.hash_lists[1].media_hashes[1].path == "BB2.txt"
    assert bb_history.hash_lists[1].media_hashes[1].hash_entries[0].action == "original"

    # the other histories don't have a new generation
    assert not os.path.isfile("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500.mhl")
    assert aa_history.latest_generation_number() == 1


def test_hash_list_name_parsing():
    def _test_regex(filename, should_match):
        filename_no_extension, _ = os.path.splitext(filename)
        if filename.endswith(ascmhl_file_extension):
            parts = re.findall(MHLHistory.history_file_name_regex, filename_no_extension)
            if should_match:
                assert len(parts) == 1 and len(parts[0]) == 2
            else:
                assert len(parts) == 0
        else:
            assert not should_match

    _test_regex("0001_AA_2020-01-16_091500.mhl", True)
    _test_regex("10001_AA_2020-01-16_091500.mhl", True)
    _test_regex("0002.mhl", True)
    _test_regex("0004_myCustomString_123.mhl", True)

    _test_regex("001_AA_2020-01-16_091500.mhl", False)
    _test_regex("0001_AA_2020-01-16_091500.xml", False)
    _test_regex("AA_2020-01-16_091500_0002.mhl", False)
    _test_regex("0003_.mhl", False)



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""
import os
import textwrap
from timeit import default_timer as timer
import dateutil.parser

from lxml import etree
from lxml.builder import E

from . import logger
from .hashlist import *
from .utils import datetime_isostring
from .__version__ import ascmhl_supported_hashformats
from .hashlist import (
    MHLCreatorInfo,
    MHLHashEntry,
    MHLHashList,
    MHLHashListReference,
    MHLMediaHash,
    MHLProcessInfo,
    MHLTool,
)
from .ignore import MHLIgnoreSpec
from .utils import datetime_isostring


def parse(file_path):
    """parsing the MHL XML file and building the MHLHashList for the hash_list member variable"""
    logger.debug(f"parsing {file_path}...")

    start = timer()

    hash_list = MHLHashList()
    hash_list.file_path = file_path
    object_stack = []
    current_object = None
    is_directory_structure = False

    # use iterparse to prevent large memory usage when parsing large files
    # pass a file handle to iterparse instead of the path directly to support the fake filesystem used in the tests
    file = open(file_path, "rb")
    existing_ignore_patterns = []
    for event, element in etree.iterparse(file, events=("start", "end")):

        # check if we need to create a new container
        if event == "start":
            # the tag might contain the namespace like {urn:ASC:MHL:v2.0}hash, so we need to strip the namespace part
            # doing it with split is faster than using the lxml QName method
            tag = element.tag.split("}", 1)[-1]

            if not current_object:
                if tag == "creatorinfo":
                    current_object = MHLCreatorInfo()
                elif tag == "processinfo":
                    current_object = MHLProcessInfo()
                elif tag == "hash":
                    current_object = MHLMediaHash()
                elif tag == "directoryhash":
                    current_object = MHLMediaHash()
                    current_object.is_directory = True
                elif tag == "hashlistreference":
                    current_object = MHLHashListReference()

            # these are the only cases where we push to the object stack
            if type(current_object) is MHLProcessInfo:
                if tag == "ignore":
                    object_stack.append(current_object)
                    current_object = MHLIgnoreSpec()
                elif tag == "roothash":
                    object_stack.append(current_object)
                    current_object = MHLMediaHash()
                    current_object.is_directory = True

            elif type(current_object) is MHLCreatorInfo:
                if tag == "author":
                    current_object.authors.append(MHLAuthor("-"))

            # take a note where we are ina <directoryhash>
            elif type(current_object) is MHLMediaHash:
                if tag == "structure":
                    is_directory_structure = True
                elif tag == "content":
                    is_directory_structure = False

        elif event == "end":

            if current_object:
                tag = element.tag.split("}", 1)[-1]

                if type(current_object) is MHLCreatorInfo:
                    if tag == "creationdate":
                        current_object.creation_date = element.text
                    elif tag == "tool":
                        current_object.tool = MHLTool(element.text, element.attrib["version"])
                    elif tag == "hostname":
                        current_object.host_name = element.text
                    elif tag == "location":
                        current_object.location = element.text
                    elif tag == "comment":
                        current_object.comment = element.text
                    elif tag == "creatorinfo":
                        hash_list.creator_info = current_object
                        current_object = None

                    # author
                    elif tag == "name":
                        current_object.authors[-1].name = element.text
                    elif tag == "role":
                        current_object.authors[-1].role = element.text
                    elif tag == "email":
                        current_object.authors[-1].email = element.text
                    elif tag == "phone":
                        current_object.authors[-1].phone = element.text

                elif type(current_object) is MHLProcessInfo:
                    if tag == "process":
                        current_object.process = element.text
                    elif tag == "processinfo":
                        hash_list.process_info = current_object
                        current_object = None
                elif type(current_object) is MHLIgnoreSpec:
                    if tag == "pattern":
                        existing_ignore_patterns.append(element.text)
                    elif tag == "ignore":
                        hash_list.process_info.ignore_spec = current_object
                        current_object = object_stack.pop()
                    else:
                        current_object = None

                elif type(current_object) is MHLMediaHash:
                    if tag == "path":
                        current_object.path = element.text
                        file_size = element.attrib.get("size")
                        current_object.file_size = int(file_size) if file_size else None
                    # TODO: parse date
                    # elif tag == 'lastmodificationdate':
                    # 	current_object.file_size = element.text
                    elif tag in ascmhl_supported_hashformats:
                        hash_date = None
                        hash_date_string = element.attrib.get("hashdate")
                        if hash_date_string is not None:
                            hash_date = dateutil.parser.parse(hash_date_string)
                        if current_object.is_directory:
                            if is_directory_structure == False:
                                entry = MHLHashEntry(tag, element.text, element.attrib.get("action"), hash_date)
                                current_object.append_hash_entry(entry)
                            else:
                                # find right hash entry and set structure hash
                                entry = current_object.find_hash_entry_for_format(tag)
                                entry.structure_hash_string = element.text
                        else:
                            entry = MHLHashEntry(tag, element.text, element.attrib.get("action"), hash_date)
                            current_object.append_hash_entry(entry)

                    elif tag == "hash" or tag == "directoryhash":
                        hash_list.append_hash(current_object)
                        current_object = None

                    elif tag == "roothash":
                        root_media_hash = current_object
                        root_media_hash.is_directory = True
                        current_object = object_stack.pop()
                        current_object.root_media_hash = root_media_hash

                elif type(current_object) is MHLHashListReference:
                    if tag == "path":
                        current_object.path = element.text
                    elif tag == "c4":
                        current_object.reference_hash = element.text
                    elif tag == "hashlistreference":
                        hash_list.append_hash_list_reference(current_object)
                        current_object = None

            # in order to keep memory usage low while parsing, we clear the finished element
            # and remove it from the parent element as well but since this is clearing the children anyways
            # we only need to do it if we are not currently parsing a container object
            if not current_object:
                element.clear()
                while element.getprevious() is not None:
                    del element.getparent()[0]

    hash_list.process_info.ignore_spec = MHLIgnoreSpec(existing_ignore_patterns)
    logger.debug(f"parsing took: {timer() - start}")

    return hash_list


def write_hash_list(hash_list: MHLHashList, file_path: str):
    """creates a new mhl file and writes the xml to disk

    we write the file step by step to reduce memory load while writing large files
    e.g. we create xml objects only for single elemnts like one media hash element and write it to disk
    before creating the next one"""

    directory_path = os.path.dirname(file_path)
    if not os.path.isdir(directory_path):
        os.mkdir(directory_path)

    file = open(file_path, "wb")
    file.write(b'<?xml version="1.0" encoding="UTF-8"?>\n<hashlist version="2.0" xmlns="urn:ASC:MHL:v2.0">\n')
    current_indent = "  "

    # set the file name early so we can use it to e.g. get the root path
    hash_list.file_path = file_path
    # write creator info
    _write_xml_element_to_file(file, _creator_info_xml_element(hash_list), "  ")
    # write process info
    _write_xml_element_to_file(file, _process_info_xml_element(hash_list), "  ")

    # write hashes
    hashes_tag = "<hashes>\n"
    _write_xml_string_to_file(file, hashes_tag, current_indent)
    current_indent += "  "

    for media_hash in hash_list.media_hashes:
        if media_hash.is_directory:
            _write_xml_element_to_file(file, _directory_hash_xml_element(media_hash), current_indent)
        else:
            _write_xml_element_to_file(file, _media_hash_xml_element(media_hash), current_indent)

    current_indent = current_indent[:-2]
    _write_xml_string_to_file(file, "</hashes>\n", current_indent)

    # only write the optional references section if there are actually some references
    if len(hash_list.referenced_hash_lists) > 0:
        _write_xml_string_to_file(file, "<references>\n", current_indent)
        current_indent += "  "
        for ref_hash_list in hash_list.referenced_hash_lists:
            _write_xml_element_to_file(file, _ascmhlreference_xml_element(ref_hash_list, file_path), current_indent)
        current_indent = current_indent[:-2]
        _write_xml_string_to_file(file, "</references>\n", current_indent)

    current_indent = current_indent[:-2]
    _write_xml_string_to_file(file, "</hashlist>\n", current_indent)
    file.flush()


def _write_xml_element_to_file(file, xml_element, indent: str):
    xml_string = etree.tostring(xml_element, pretty_print=True, encoding="unicode")
    _write_xml_string_to_file(file, xml_string, indent)


def _write_xml_string_to_file(file, xml_string: str, indent: str):
    result = textwrap.indent(xml_string, indent)
    file.write(result.encode("utf-8"))


def _media_hash_xml_element(media_hash: MHLMediaHash):
    """builds and returns one <hash> element for a given MediaHash object"""

    path_element = E.path(media_hash.path)
    if media_hash.file_size:
        path_element.attrib["size"] = str(media_hash.file_size)
    if media_hash.last_modification_date:
        path_element.attrib["lastmodificationdate"] = datetime_isostring(media_hash.last_modification_date)

    hash_element = E.hash(path_element)
    sorted_hash_entries = sorted(media_hash.hash_entries, key=lambda hash_entry: hash_entry.hash_format)
    for hash_entry in sorted_hash_entries:
        entry_element = E(hash_entry.hash_format)
        entry_element.text = hash_entry.hash_string
        if hash_entry.action:
            entry_element.attrib["action"] = hash_entry.action
        if hash_entry.hash_date:
            entry_element.attrib["hashdate"] = datetime_isostring(hash_entry.hash_date, True)
        hash_element.append(entry_element)

    return hash_element


def _directory_hash_xml_element(media_hash: MHLMediaHash, skipPath=False):
    """builds and returns one <hash> element for a given MediaHash object"""

    content_element = E.content()
    structure_element = E.structure()

    for hash_entry in media_hash.hash_entries:
        entry_element_content = E(hash_entry.hash_format)
        entry_element_content.text = hash_entry.hash_string

        entry_element_structure = E(hash_entry.hash_format)
        entry_element_structure.text = hash_entry.structure_hash_string

        if hash_entry.action:
            entry_element_content.attrib["action"] = hash_entry.action
            entry_element_structure.attrib["action"] = hash_entry.action
        if hash_entry.hash_date:
            entry_element_content.attrib["hashdate"] = datetime_isostring(hash_entry.hash_date, True)
            entry_element_structure.attrib["hashdate"] = datetime_isostring(hash_entry.hash_date, True)

        content_element.append(entry_element_content)
        structure_element.append(entry_element_structure)

    hash_element = E.directoryhash()

    if skipPath == False:
        path_element = E.path(media_hash.path)
        if media_hash.file_size:
            path_element.attrib["size"] = str(media_hash.file_size)
        if media_hash.last_modification_date:
            path_element.attrib["lastmodificationdate"] = datetime_isostring(media_hash.last_modification_date)
        hash_element.append(path_element)

    hash_element.append(content_element)
    hash_element.append(structure_element)

    return hash_element


def _ascmhlreference_xml_element(hash_list: MHLHashList, file_path: str):
    """builds and returns one <hashlistreference> element for a given HashList object"""

    root_path = os.path.dirname(os.path.dirname(file_path))
    hash_element = E.hashlistreference(
        E.path(os.path.relpath(hash_list.file_path, root_path)),
        E.c4(hash_list.generate_reference_hash()),
    )

    return hash_element


def _creator_info_xml_element(hash_list: MHLHashList):
    """builds and returns one <creatorinfo> element for a given creator info instance"""
    creator_info = hash_list.creator_info
    info_element = E.creatorinfo(
        E.creationdate(creator_info.creation_date),
        E.hostname(creator_info.host_name),
        E.tool(creator_info.tool.name, version=creator_info.tool.version),
    )
    if creator_info.location is not None:
        info_element.append(E.location(creator_info.location))

    if creator_info.comment is not None:
        info_element.append(E.comment(creator_info.comment))

    for author in creator_info.authors:
        author_element = E.author()
        if author.name is not None:
            author_element.append(E.name(author.name))
        if author.role is not None:
            author_element.append(E.role(author.role))
        if author.email is not None:
            author_element.append(E.email(author.email))
        if author.phone is not None:
            author_element.append(E.phone(author.phone))
        info_element.append(author_element)

    return info_element


def _process_info_xml_element(hash_list: MHLHashList):
    """builds and returns one <creatorinfo> element for a given creator info instance"""
    process_info = hash_list.process_info
    # create empty root hash if directory hashes are disabled or use the generated one from the hash list
    root_hash = hash_list.process_info.root_media_hash
    if not root_hash:
        root_hash = MHLMediaHash()
        root_hash.path = hash_list.get_root_path()
    if root_hash.path == ".":  # TODO: can we find out if this is a non-flattened history further up?
        root_hash.path = hash_list.get_root_path()

    info_element = E.processinfo(
        E.process(process_info.process.process_type),
        _root_media_hash_xml_element(root_hash),
        _ignorespec_xml_element(hash_list.process_info.ignore_spec),
    )
    return info_element


def _ignorespec_xml_element(ignore_spec: MHLIgnoreSpec):
    spec_element = E.ignore()
    if ignore_spec:
        for ignore_pattern in ignore_spec.get_pattern_list():
            spec_element.append(E.pattern(ignore_pattern))
    return spec_element


def _ignore_xml_element(ignore_pattern: str):
    """builds and returns one <ignore> element for a given ignore_spec pattern entry"""
    return E.ignore(ignore_pattern)


def _root_media_hash_xml_element(root_media_hash: MHLMediaHash):
    element = _directory_hash_xml_element(root_media_hash, skipPath=True)
    element.tag = "roothash"
    return element



#!/usr/bin/env python3
"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import click

# noinspection PyProtectedMember
from ascmhl import _debug_commands


@click.group()
def debug_cli():
    pass


debug_cli.add_command(_debug_commands.readmhlfile)
debug_cli.add_command(_debug_commands.readchainfile)
debug_cli.add_command(_debug_commands.readmhlhistory)
debug_cli.add_command(_debug_commands.create_dummy_file_structure, "create_dummy_file_structure")


if __name__ == "__main__":
    debug_cli()



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""
from typing import List

from . import logger


class MHLChain:
    """
    class for representing a chain file with its list of generations (one for each ascmhl file)
    managed by the MHLHistory class
    uses MHLChainGeneration for storing information

    model member variables:
    mhl_history -- MHLHistory object for context
    generations -- list of MHLChainGeneration objects

    attribute member variables:
    file path -- absolute path to chain file
    """

    # init

    file_path: str
    generations: List["MHLChainGeneration"]

    def __init__(self, file_path: str):
        self.file_path = file_path
        self.generations = []

    # build

    def append_generation(self, generation):
        self.generations.append(generation)

    # log

    def log(self):
        logger.info("     file path: {0}".format(self.file_path))
        for generation in self.generations:
            generation.log()


class MHLChainGeneration:
    """
    class for representing one generation
    managed by a MHLChain object

    model member variables:
    mhl_chain -- MHLChain object for context

    attribute member variables:
    generation_number -- integer, -1 means invalid
    ascmhl_filename --
    hashformat --
    hash_string --

    other member variables:
    """

    generation_number: int  # -1 means invalid
    ascmhl_filename: str
    hash_format: str
    hash_string: str

    def __init__(self, generation_number=-1, ascmhl_filename=None, hash_format=None, hash_string=None):
        # line string examples:
        # 0001 A002R2EC_2019-10-08_100916_0001.ascmhl SHA1: 9e9302b3d7572868859376f0e5802b87bab3199e

        self.generation_number = generation_number
        self.ascmhl_filename = ascmhl_filename
        self.hash_format = hash_format
        self.hash_string = hash_string

    def log(self):
        action = "*"  # FIXME
        indicator = " "

        logger.info(
            "{0} {1}: {2} {3}: {4}".format(
                indicator,
                self.hash_format.rjust(6),
                self.hash_string.ljust(32),
                action.ljust(10),
                self.ascmhl_filename,
            )
        )



"""
__author__ = "Jon Waggoner"
__copyright__ = "Copyright 2020, Pomfort GmbH"
__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
import pytest
import shutil
from click.testing import CliRunner
from lxml import etree
from testfixtures import TempDirectory
import ascmhl.commands


XML_IGNORE_TAG = "pattern"
XML_HASH_PATH_TAG = "path"

DEFAULT_IGNORE_SET = {".DS_Store", "ascmhl", "ascmhl/"}


@pytest.fixture(scope="function")
def temp_tree():
    """
    fixture used by all tests in this file.
    sets up a static file tree in a temp dir for each test, then deletes after test.
    """
    # create and populate a temp dir with static data
    tmpdir = TempDirectory()
    tmpdir.write("a.txt", b"a")
    tmpdir.write("b.txt", b"b")
    tmpdir.write("1/c.txt", b"c")
    tmpdir.write("1/11/d.txt", b"d")
    tmpdir.write("1/12/e.txt", b"e")
    tmpdir.write("2/11/f.txt", b"f")
    tmpdir.write("2/12/g.txt", b"g")

    # yield the directory to the test function
    yield tmpdir

    # test function is done... cleanup the temp dir
    try:
        tmpdir.cleanup()
    except OSError as e:
        print("Error: %s - %s." % (e.filename, e.strerror))


def ignore_patterns_from_mhl_file(mhl_file: str):
    """
    returns a set of the patterns found in the mhl_file
    """
    pattern_list = [
        element.text for event, element in etree.iterparse(mhl_file) if element.tag.split("}", 1)[-1] == XML_IGNORE_TAG
    ]
    return set(pattern_list)


def paths_from_mhl_file(mhl_file: str):
    """
    returns a set of the hash entires found in the mhl_file
    """
    hash_list = [
        element.text
        for event, element in etree.iterparse(mhl_file)
        if element.tag.split("}", 1)[-1] == XML_HASH_PATH_TAG
    ]
    return set(hash_list)


def assert_mhl_file_has_exact_ignore_patterns(mhl_file: str, patterns_to_check: set):
    """
    asserts the ignore patterns in an mhl are exactly the same as the patterns in patterns_to_check
    this function adds the default ignore spec to the patterns to check since they are always expected
    """
    patterns_in_file = set(ignore_patterns_from_mhl_file(mhl_file))
    assert patterns_in_file == DEFAULT_IGNORE_SET | patterns_to_check, "mhl file has incorrect ignore patterns"


def mhl_file_for_gen(mhl_dir: str, mhl_gen: int):
    """
    returns the mhl file associated with a generation number.
    """
    assert mhl_gen != 0  # ensure we sent the gen number and not array number
    mhl_files = os.listdir(mhl_dir)
    mhl_files.sort()
    return f"{mhl_dir}/{mhl_files[mhl_gen - 1]}"


def test_default_ignore(temp_tree):
    """
    tests that .DS_Store and ascmhl/ fs entries are ignored from the mhl spec
    """
    runner = CliRunner()
    root_dir, mhl_dir = f"{temp_tree.path}", f"{temp_tree.path}/ascmhl"

    # write 2 generations (since ascmhl folder won't exist til after gen 1)
    assert runner.invoke(ascmhl.commands.create, [root_dir]).exit_code == 0
    assert runner.invoke(ascmhl.commands.create, [root_dir]).exit_code == 0

    # ensure we have the expected content in the ascmhl folder
    gen_2_file = mhl_file_for_gen(mhl_dir, 2)
    assert gen_2_file

    gen_2_paths = paths_from_mhl_file(gen_2_file)
    assert "a.txt" in gen_2_paths
    assert "ascmhl" not in gen_2_paths


def test_ignore_on_create(temp_tree):
    """
    tests that the "create" command properly receives and processes all ignore spec arguments
    """
    runner = CliRunner()
    root_dir, mhl_dir = f"{temp_tree.path}", f"{temp_tree.path}/ascmhl"

    # write generation 1,
    runner.invoke(ascmhl.commands.create, [root_dir, "-i", "1"])
    # write generation 2, appending ignore patterns using both CLI args
    runner.invoke(ascmhl.commands.create, [root_dir, "-i", "2", "--ignore", "3"])
    # write generation 3, appending ignore_spec from file and both CLI args
    temp_tree.write("ignorespec", b"6\n7")  # write an ignore spec to file
    runner.invoke(
        ascmhl.commands.create, [root_dir, "-i", "4", "--ignore", "5", "--ignore_spec", f"{temp_tree.path}/ignorespec"]
    )

    # we should now have 3 total mhl generations. ensure each one has exactly the expected patterns
    mhl_files = os.listdir(f"{temp_tree.path}/ascmhl")
    mhl_files.sort()
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(mhl_dir, 1), {"1"})
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(mhl_dir, 2), {"1", "2", "3"})
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(mhl_dir, 3), {"1", "2", "3", "4", "5", "6", "7"})


def test_ignore_on_verify(temp_tree):
    """
    tests that the "verify" command properly receives and processes all ignore spec arguments
    """
    runner = CliRunner()
    root_dir, mhl_dir = f"{temp_tree.path}", f"{temp_tree.path}/ascmhl"

    # create a generation
    runner.invoke(ascmhl.commands.create, [root_dir])
    # purposely alter integrity by deleting a file
    os.remove(f"{root_dir}/a.txt")
    # assert that verification fails
    assert runner.invoke(ascmhl.commands.verify, [root_dir]).exit_code != 0
    # assert that we can suppress this verification error by ignoring the deleted file
    assert runner.invoke(ascmhl.commands.verify, [root_dir, "-i", "a.txt"]).exit_code == 0

    # again, but by ignoring a dir. alter a file, then ignore the parent.
    temp_tree.write(f"{root_dir}/1/c.txt", b"BAD_CONTENTS")
    # assert that verification fails
    assert runner.invoke(ascmhl.commands.verify, [root_dir]).exit_code != 0
    # assert that we can suppress this verification error by ignoring a parent directory
    assert runner.invoke(ascmhl.commands.verify, [root_dir, "-i", "a.txt", "--ignore", "c.txt"]).exit_code == 0

    # assert that the same verification works by using an ignore spec from file
    temp_tree.write("ignorespec", b"a.txt\n1/c.txt")  # write an ignore spec to file
    assert runner.invoke(ascmhl.commands.verify, [root_dir, "--ignore_spec", f"{temp_tree.path}/ignorespec"])


def test_ignore_on_diff(temp_tree):
    """
    tests that the "diff" command properly receives and processes all ignore spec arguments
    """
    runner = CliRunner()
    root_dir, mhl_dir = f"{temp_tree.path}", f"{temp_tree.path}/ascmhl"

    # create a generation
    runner.invoke(ascmhl.commands.create, [root_dir])
    # purposely alter integrity by deleting a file
    os.remove(f"{root_dir}/a.txt")
    # assert that verification fails
    assert runner.invoke(ascmhl.commands.diff, [root_dir]).exit_code != 0
    # assert that we can suppress this verification error by ignoring the deleted file
    assert runner.invoke(ascmhl.commands.diff, [root_dir, "-i", "a.txt"]).exit_code == 0

    # again, but by ignoring a dir. alter a file, then ignore the parent.
    temp_tree.write(f"{root_dir}/1/c.txt", b"BAD_CONTENTS")
    # assert that verification fails
    assert runner.invoke(ascmhl.commands.diff, [root_dir]).exit_code != 0
    # assert that we can suppress this verification error by ignoring a parent directory
    assert runner.invoke(ascmhl.commands.diff, [root_dir, "-i", "a.txt", "--ignore", "1/c.txt"]).exit_code == 0

    # assert that the same verification works by using an ignore spec from file
    temp_tree.write("ignorespec", b"a.txt\n1/c.txt")  # write an ignore spec to file
    assert runner.invoke(ascmhl.commands.diff, [root_dir, "--ignore_spec", f"{temp_tree.path}/ignorespec"])


def test_ignore_on_nested_histories(temp_tree):
    """
    tests that nested histories have proper interaction with the ignore specification.
    parent ignore specs are propagated to their children and appended to the existing child specs.
    ensure we append but do not overwrite.
    """
    runner = CliRunner()
    root_dir, root_mhl_dir = f"{temp_tree.path}", f"{temp_tree.path}/ascmhl"
    child_dir, child_mhl_dir = f"{temp_tree.path}/1", f"{temp_tree.path}/1/ascmhl"

    # create an mhl gen on a nested dir
    runner.invoke(ascmhl.commands.create, [child_dir, "-i", "c1"])
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(child_mhl_dir, 1), {"c1"})

    # create mhl gen on parent, ensure parent is correct and child updated.
    runner.invoke(ascmhl.commands.create, [root_dir, "-i", "p1"])
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(root_mhl_dir, 1), {"p1"})
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(child_mhl_dir, 2), {"c1", "p1"})

    # add just to child again.
    runner.invoke(ascmhl.commands.create, [child_dir, "-i", "c2"])
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(child_mhl_dir, 3), {"c1", "p1", "c2"})

    # add to parent and test one last time.
    runner.invoke(ascmhl.commands.create, [root_dir, "-i", "p2"])
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(root_mhl_dir, 2), {"p1", "p2"})
    assert_mhl_file_has_exact_ignore_patterns(mhl_file_for_gen(child_mhl_dir, 4), {"c1", "p1", "c2", "p2"})



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""
try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version  # noqa

ascmhl_tool_name = "ascmhl"
ascmhl_tool_version = version("ascmhl")

ascmhl_folder_name = "ascmhl"
ascmhl_file_extension = ".mhl"
ascmhl_chainfile_name = "ascmhl_chain.xml"
ascmhl_collectionfile_name = "ascmhl_collection.xml"
# decreasing priority list for verification
ascmhl_supported_hashformats = [
    "md5",
    "sha1",
    "xxh128",
    "xxh3",
    "xxh64",
    "c4",
]
ascmhl_default_hashformat = "xxh128"
ascmhl_reference_hash_format = "c4"  # hash format used to reference other files, e.g. in references and the chain
# ascmhl_default_ignore_patterns = ['.DS_Store', ascmhl_folder_name]



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import datetime
import time


def matches_prefixes(text: str, prefixes: list):
    for prefix in prefixes:
        if text.startswith(prefix):
            return True
    return False


def datetime_isostring(date, keep_microseconds=False):
    """create an iso string representation for a date object
    e.g. for use in XML tags and attributes

    arguments:
    date -- date object
    keep_microseconds -- include microseconds in iso
    """
    utc_offset_sec = time.altzone if time.localtime().tm_isdst else time.timezone
    utc_offset = datetime.timedelta(seconds=-utc_offset_sec)

    if keep_microseconds:
        date_to_format = date
    else:
        date_to_format = date.replace(microsecond=0)

    return date_to_format.replace(tzinfo=datetime.timezone(offset=utc_offset)).isoformat()


def datetime_now_isostring():
    return datetime_isostring(datetime.datetime.now())


def datetime_now_filename_string():
    """create a string representation for now() for use as part of the MHL filename"""
    return datetime.datetime.strftime(datetime.datetime.utcnow(), "%Y-%m-%d_%H%M%S")


def datetime_now_isostring_with_microseconds():
    return datetime_isostring(datetime.datetime.now(), keep_microseconds=True)



from packaging import version
from threading import Thread

import requests

from ascmhl.__version__ import ascmhl_tool_version


class Updater(Thread):
    def __init__(self):
        super().__init__()
        self.daemon = True
        self.latest_version = None
        self.finished = False
        self.start()

    def run(self):
        self._get_latest_version()
        self.finished = True

    @property
    def needs_update(self) -> bool:
        if not self.latest_version:
            return False

        return (
            self.latest_version > version.parse(ascmhl_tool_version)
            and not self.latest_version.is_devrelease
            and not self.latest_version.is_prerelease
        )

    def _get_latest_version(self):
        try:
            r = requests.get("https://api.github.com/repos/ascmitc/mhl/releases/latest")
            r.raise_for_status()
            self.latest_version = version.parse(r.json().get("tag_name"))
        except requests.exceptions.RequestException:
            self.finished = True



"""
__author__ = "Jon Waggoner, Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner"
__email__ = "opensource@pomfort.com"
"""

import click


class CompletenessCheckFailedException(click.ClickException):
    exit_code = 15

    def __init__(self):
        super().__init__("Files referenced in the ASC MHL history are missing")


class NoMHLHistoryException(click.ClickException):
    exit_code = 11

    def __init__(self, path):
        super().__init__(f"Missing ASC MHL history at path {path}")


class VerificationFailedException(click.ClickException):
    exit_code = 12

    def __init__(self):
        super().__init__("Verification of files referenced in the ASC MHL history failed")


class VerificationDirectoriesFailedException(click.ClickException):
    exit_code = 15

    def __init__(self):
        super().__init__("Verification of directories referenced in the ASC MHL history failed")


class NewFilesFoundException(click.ClickException):
    exit_code = 13

    def __init__(self):
        super().__init__("New files not referenced in the ASC MHL history have been found")


class NoMHLHistoryExceptionForPath(click.ClickException):
    exit_code = 14

    def __init__(self, path):
        super().__init__(f"Missing ASC MHL history for path {path}")


class SingelFileNotFoundException(click.ClickException):
    exit_code = 15

    def __init__(self):
        super().__init__("Could not find single file.")



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from collections import defaultdict
from typing import Dict, List

from . import chain_xml_parser
from . import logger
from .ignore import MHLIgnoreSpec
from .hashlist import MHLHashList, MHLHashEntry, MHLCreatorInfo, MHLProcessInfo
from .history import MHLHistory


class MHLGenerationCreationSession:
    """
    class for representing a verification session

    the session is based on a MHLHistory object (that can be empty if no as-mhl folder exists yet).
    It is used to document the verification of single files and will handle verifying the given hash against
    existing generations. While being used by the verification process it will create a new generation (hash list)
    and write it to disk once the session is committed.

    - public interface
        * initialized with a MHLHistory object
        * adding hashes generated from certain files
        * committing of a session to write the new generation to disk
    """

    root_history: MHLHistory
    new_hash_lists: Dict[MHLHistory, MHLHashList]
    ignore_spec: MHLIgnoreSpec

    def __init__(self, history: MHLHistory, ignore_spec: MHLIgnoreSpec = MHLIgnoreSpec()):
        self.root_history = history
        self.new_hash_lists = defaultdict(MHLHashList)
        self.ignore_spec = ignore_spec

    def append_file_hash(
        self, file_path, file_size, file_modification_date, hash_format, hash_string, action=None, hash_date=None
    ) -> bool:

        relative_path = self.root_history.get_relative_file_path(file_path)
        # TODO: handle if path is outside of history root path

        history, history_relative_path = self.root_history.find_history_for_path(relative_path)
        # for collections we cannot create a valid relative path (we are in the "wrong" history), but in that case
        # the file_path is inputted already as the relative path (a bit of implicit functionality here)
        if history_relative_path == None:
            history_relative_path = file_path

        # check if there is an existing hash in the other generations and verify
        original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

        hash_entry = MHLHashEntry(hash_format, hash_string, hash_date=hash_date)
        if original_hash_entry is None:
            hash_entry.action = "original"
            logger.verbose(f"  created original hash for     {relative_path}  {hash_format}: {hash_string}")
        else:
            existing_hash_entry = history.find_first_hash_entry_for_path(history_relative_path, hash_format)
            if existing_hash_entry is not None:
                if existing_hash_entry.hash_string == hash_string:
                    hash_entry.action = "verified"
                    logger.verbose(f"  verified                      {relative_path}  OK")
                else:
                    hash_entry.action = "failed"
                    logger.error(
                        f"ERROR: hash mismatch for        {relative_path}  "
                        f"{hash_format} (old): {existing_hash_entry.hash_string}, "
                        f"{hash_format} (new): {hash_string}"
                    )
            else:
                # in case there is no hash entry for this hash format yet
                hash_entry.action = "new"  # mark as 'new' here, will be changed to verified in _validate_new_hash_list
                logger.verbose(
                    f"  created new, verified hash for          {relative_path}  {hash_format}: {hash_string}"
                )

        # in case the same file is hashes multiple times we want to add all hash entries
        new_hash_list = self.new_hash_lists[history]
        media_hash = new_hash_list.find_or_create_media_hash_for_path(
            history_relative_path, file_size, file_modification_date
        )

        # collection behavior: overwrite action with action from flattened history
        if action != None:
            hash_entry.action = action

        media_hash.append_hash_entry(hash_entry)
        return hash_entry.action != "failed"

    def append_directory_hashes(
        self, path, modification_date, hash_format, content_hash_string, structure_hash_string
    ) -> None:

        relative_path = self.root_history.get_relative_file_path(path)
        # TODO: handle if path is outside of history root path

        history, history_relative_path = self.root_history.find_history_for_path(relative_path)

        # in case the same file is hashes multiple times we want to add all hash entries
        new_hash_list = self.new_hash_lists[history]
        media_hash = new_hash_list.find_or_create_media_hash_for_path(history_relative_path, None, modification_date)
        media_hash.is_directory = True

        if content_hash_string:
            hash_entry = MHLHashEntry(hash_format, content_hash_string)
            hash_entry.structure_hash_string = structure_hash_string
            media_hash.append_hash_entry(hash_entry)
            if relative_path == ".":
                logger.verbose(
                    f"  calculated root hash  {hash_format}: "
                    f"{content_hash_string} (content), "
                    f"{structure_hash_string} (structure)"
                )
            else:
                logger.verbose(
                    f"  calculated directory hash for {relative_path}  {hash_format}: "
                    f"{content_hash_string} (content), "
                    f"{structure_hash_string} (structure)"
                )
        else:
            logger.verbose(f"  added directory entry for     {relative_path}")

        # in case we just created the root media hash of the current hash list we also add it one history level above
        if new_hash_list.process_info.root_media_hash is media_hash and history.parent_history:
            parent_history = history.parent_history
            parent_relative_path = parent_history.get_relative_file_path(path)
            parent_hash_list = self.new_hash_lists[parent_history]
            parent_media_hash = parent_hash_list.find_or_create_media_hash_for_path(
                parent_relative_path, None, modification_date
            )
            parent_media_hash.is_directory = True
            if content_hash_string:
                hash_entry = MHLHashEntry(hash_format, content_hash_string)
                hash_entry.structure_hash_string = structure_hash_string
                parent_media_hash.append_hash_entry(hash_entry)

    def commit(self, creator_info: MHLCreatorInfo, process_info: MHLProcessInfo):
        """
        this method needs to create the generations of the children bottom up
        # so each history can reference the children correctly and can get the actual hash of the file
        """

        # store all references to child histories we will need when committing the parent history
        referenced_hash_lists: Dict[MHLHistory, List[MHLHashList]] = defaultdict(list)

        for history in MHLHistory.walk_child_histories(self.root_history):
            if history not in self.new_hash_lists:
                if history not in referenced_hash_lists:
                    continue
                new_hash_list = MHLHashList()
            else:
                new_hash_list = self.new_hash_lists[history]
            new_hash_list.referenced_hash_lists = referenced_hash_lists[history]
            new_hash_list.creator_info = creator_info

            # only for flattening we want to inject a custom root hash
            if not new_hash_list.process_info.root_media_hash:
                new_hash_list.process_info.root_media_hash = process_info.root_media_hash
            new_hash_list.process_info.hashlist_custom_basename = process_info.hashlist_custom_basename
            new_hash_list.process_info.process = process_info.process
            new_hash_list.process_info.ignore_spec = MHLIgnoreSpec(
                history.latest_ignore_patterns(), self.ignore_spec.get_pattern_list()
            )

            history.write_new_generation(new_hash_list)
            relative_generation_path = self.root_history.get_relative_file_path(new_hash_list.file_path)
            logger.verbose(f"Created new generation {relative_generation_path}")
            if history.parent_history is not None:
                referenced_hash_lists[history.parent_history].append(new_hash_list)

            chain_xml_parser.write_chain(history.chain, new_hash_list)



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from click.testing import CliRunner
from freezegun import freeze_time

import ascmhl.commands


@freeze_time("2020-01-16 09:15:00")
def test_simple_info_fails_no_history(fs, simple_mhl_history):
    runner = CliRunner()
    os.rename("/root/ascmhl", "/root/_ascmhl")
    result = runner.invoke(ascmhl.commands.info, ["-sf", "/root/Stuff.txt"])
    assert result.exit_code == 14


@freeze_time("2020-01-16 09:15:00")
def test_simple_info(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.info, ["-sf", "/root/Stuff.txt"])
    assert (
        result.output
        == "Info with history at path: /root\nStuff.txt:\n  Generation 1 (2020-01-15T13:00:00+00:00) xxh64:"
        " 94c399c2a9a21f9a (original)\n"
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_info(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, ["/root/", "-h", "xxh64"])
    result = runner.invoke(ascmhl.commands.create, ["/root/", "-h", "md5"])
    result = runner.invoke(ascmhl.commands.create, ["/root/", "-h", "xxh64"])
    result = runner.invoke(ascmhl.commands.info, ["-sf", "/root/Stuff.txt"])
    assert (
        result.output
        == "Info with history at path: /root\nStuff.txt:\n"
        "  Generation 1 (2020-01-15T13:00:00+00:00) xxh64: 94c399c2a9a21f9a (original)\n"
        "  Generation 2 (2020-01-16T09:15:00+00:00) xxh64: 94c399c2a9a21f9a (verified)\n"
        "  Generation 3 (2020-01-16T09:15:00+00:00) md5: 9eb84090956c484e32cb6c08455a667b (verified)\n"
        "  Generation 3 (2020-01-16T09:15:00+00:00) xxh64: 94c399c2a9a21f9a (verified)\n"
        "  Generation 4 (2020-01-16T09:15:00+00:00) xxh64: 94c399c2a9a21f9a (verified)\n"
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_altered_file(fs, simple_mhl_history):
    # alter a file
    with open("/root/Stuff.txt", "a") as file:
        file.write("!!")
    CliRunner().invoke(ascmhl.commands.create, ["/root"])
    CliRunner().invoke(ascmhl.commands.create, ["/root", "-h", "md5"])

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.info, ["-sf", "/root/Stuff.txt"])
    assert (
        result.output
        == "Info with history at path: /root\nStuff.txt:\n  Generation 1 (2020-01-15T13:00:00+00:00) xxh64:"
        " 94c399c2a9a21f9a (original)\n  Generation 2 (2020-01-16T09:15:00+00:00) xxh64: 2346e97eb08788cc (failed)\n"
        "  Generation 3 (2020-01-16T09:15:00+00:00) xxh64: 2346e97eb08788cc (failed)\n"
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_nested_info(fs, nested_mhl_histories):
    CliRunner().invoke(ascmhl.commands.create, ["/root", "-h", "xxh64"])
    CliRunner().invoke(ascmhl.commands.create, ["/root", "-h", "md5"])

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.info, ["-sf", "/root/Stuff.txt"])
    assert (
        result.output
        == "Info with history at path: /root\nStuff.txt:\n"
        "  Generation 1 (2020-01-15T13:00:00+00:00) xxh64: 94c399c2a9a21f9a (original)\n"
        "  Generation 2 (2020-01-16T09:15:00+00:00) xxh64: 94c399c2a9a21f9a (verified)\n"
        "  Generation 3 (2020-01-16T09:15:00+00:00) md5: 9eb84090956c484e32cb6c08455a667b (verified)\n"
        "  Generation 3 (2020-01-16T09:15:00+00:00) xxh64: 94c399c2a9a21f9a (verified)\n"
    )
    assert result.exit_code == 0



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from . import logger
from .__version__ import ascmhl_reference_hash_format
from .chain import MHLChain, MHLChainGeneration
from .hashlist import MHLHashList
from .hasher import hash_file
import os


def parse(file_path):
    """parsing the chain.txt file and building the MHLChain for the chain member variable"""
    logger.debug(f'parsing "{os.path.basename(file_path)}"...')

    chain = MHLChain(file_path)

    if not os.path.exists(file_path):
        return chain

    lines = [line.rstrip("\n") for line in open(file_path)]

    for line in lines:
        line = line.rstrip().lstrip()
        if line != "" and not line.startswith("#"):
            generation = _generation_from_line_in_chainfile(line)
            if generation is None:
                logger.error("cannot read line")
                continue
            chain.append_generation(generation)

    return chain


def _generation_from_line_in_chainfile(line):
    """creates a Generation object from a line int the chain file"""

    # TODO split by whitespace
    parts = line.split(None)

    if parts is not None and parts.__len__() < 4:
        logger.error('cannot read line "{line}"')
        return None

    generation = MHLChainGeneration(
        int(parts[0]),
        parts[1],
        (parts[2])[:-1],
        parts[3],
    )

    if parts.__len__() == 6:
        generation.signature_identifier = parts[4]
        generation.signature = parts[5]

    # TODO sanity checks
    return generation


def write_chain(chain: MHLChain, new_hash_list: MHLHashList):
    logger.debug(f'writing "{os.path.basename(chain.file_path)}"...')
    _append_new_generation_to_file(chain, new_hash_list)


def _line_for_chainfile(chain_generation: MHLChainGeneration):
    """creates a text line for appending a generation to a chain file"""
    result_string = (
        str(chain_generation.generation_number).zfill(4)
        + " "
        + chain_generation.ascmhl_filename
        + " "
        + chain_generation.hash_format
        + ": "
        + chain_generation.hash_string
    )

    return result_string


def _append_new_generation_to_file(chain: MHLChain, hash_list: MHLHashList):
    """appends an externally created Generation object to the chain file"""

    # get a new generation for a hashlist
    generation = MHLChainGeneration(
        hash_list.generation_number,
        hash_list.get_file_name(),
        ascmhl_reference_hash_format,
        hash_file(hash_list.file_path, ascmhl_reference_hash_format),
    )

    # TODO sanity checks
    # - if generation is already part of self.generations
    # - if generation number is sequential

    # immediately write to file
    logger.debug(f'   appending chain generation for "{generation.ascmhl_filename}" to chain file')

    with open(chain.file_path, "a") as file:
        file.write(_line_for_chainfile(generation) + "\n")

    # FIXME: check if file could be created

    # …and store here
    # FIXME: only if successfully written to file
    chain.generations.append(generation)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    author="Patrick Renner, Alexander Sahm",
    author_email="opensource@pomfort.com",
    description="ASC Media Hash List (ASC MHL)",
    entry_points={"console_scripts": ["ascmhl = ascmhl.cli.ascmhl:mhltool_cli"]},
    include_package_data=True,
    install_requires=[
        "Click>=7.0",
        "lxml>=4.6.2",
        "packaging>=20.9",
        "pathspec>=0.8.0",
        "requests>=2.25.1",
        "xxhash>=2.0.0",
        "python-dateutil>=2.8.2",
        "importlib-metadata>=4.0.1; python_version < '3.8'",
    ],
    dependency_links=[],
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="ascmhl",
    packages=find_packages(),
    setup_requires=["pytest-runner", "setuptools_scm"],
    tests_require=["pytest", "pytest_mock", "pyfakefs", "freezegun", "requests-mock", "testfixtures"],
    url="https://github.com/ascmitc/mhl",
    use_scm_version=True,
    python_requires="~=3.7",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from click.testing import CliRunner

from ascmhl.__version__ import ascmhl_tool_version
from ascmhl.cli.ascmhl import mhltool_cli


def test_version():
    runner = CliRunner()
    result = runner.invoke(mhltool_cli, "--version")
    assert result.exit_code == 0
    assert ascmhl_tool_version in result.output



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
import shutil

import click
from .history import MHLHistory
from . import chain_xml_parser
from . import hashlist_xml_parser


@click.command()
@click.argument("filepath", type=click.Path(exists=True))
@click.option("--verbose", "-v", default=False, is_flag=True, help="Verbose output")
def readchainfile(filepath, verbose):
    """
    read an ASC-MHL file
    """

    chain = chain_xml_parser.parse(filepath)

    if verbose:
        chain.log()


@click.command()
@click.argument("filepath", type=click.Path(exists=True))
@click.option("--verbose", "-v", default=False, is_flag=True, help="Verbose output")
def readmhlfile(filepath, verbose):
    """
    read an ASC-MHL file
    """

    hashlist = hashlist_xml_parser.parse(filepath)

    if verbose:
        hashlist.log()


@click.command()
@click.argument("root_path", type=click.Path(exists=True))
@click.option("--verbose", "-v", default=False, is_flag=True, help="Verbose output")
def readmhlhistory(root_path, verbose):
    """
    read an ASC-MHL file
    """

    history = MHLHistory.load_from_path(root_path)

    if verbose:
        history.log()


@click.command()
@click.argument("root_path", type=click.Path(exists=False))
def create_dummy_file_structure(root_path):
    """
    create a potentially huge set of dummy files to test how the commands can handle large file systems
    """

    folder_depth = 3
    print("delete old dummy folder")
    dummy_folder = os.path.join(root_path, "dummy_fs")
    if os.path.exists(dummy_folder):
        shutil.rmtree(dummy_folder)
    os.makedirs(dummy_folder, exist_ok=True)
    create_dummy_folder(dummy_folder, "", folder_depth)


def create_dummy_folder(root_path, prefix, depth):
    num_folders = 10
    num_files = 200
    verbose = False
    if len(prefix) > 0:
        folder_path = os.path.join(root_path, prefix)
        print(f"d: {folder_path}")
        os.mkdir(folder_path)
        for file in range(0, num_files):
            file_name = f"{prefix}{file:03}.txt"
            file_path = os.path.join(folder_path, file_name)
            if verbose:
                print(f"f: {file_path}")
            with open(file_path, "w") as file_handle:
                file_handle.write(file_name)
    else:
        folder_path = root_path
    if depth == 0:
        return
    for folder in range(0, num_folders):
        folder_name = prefix + chr(ord("A") + folder)
        create_dummy_folder(folder_path, folder_name, depth - 1)



import time
from packaging import version

import requests


def test_updater(requests_mock, mocker):
    mocker.patch("ascmhl.__version__.ascmhl_tool_version", "0.0.1")
    requests_mock.get(
        "https://api.github.com/repos/ascmitc/mhl/releases/latest",
        json={"tag_name": "v0.6.5"},
    )

    from importlib import reload
    import ascmhl.cli.update

    reload(ascmhl.cli.update)

    updater = ascmhl.cli.update.Updater()
    while not updater.finished:
        time.sleep(0.1)

    assert updater.latest_version == version.parse("0.6.5")
    assert updater.needs_update is True
    updater.join(timeout=1)


def test_updater_prerelease(requests_mock, mocker):
    mocker.patch("ascmhl.__version__.ascmhl_tool_version", "0.0.1")
    requests_mock.get(
        "https://api.github.com/repos/ascmitc/mhl/releases/latest",
        json={"tag_name": "v0.6.5-alpha.2"},
    )

    from importlib import reload
    import ascmhl.cli.update

    reload(ascmhl.cli.update)

    updater = ascmhl.cli.update.Updater()
    while not updater.finished:
        time.sleep(0.1)

    assert updater.latest_version == version.parse("0.6.5a2")
    assert updater.needs_update is False
    updater.join(timeout=1)


def test_updater_timeout(requests_mock, mocker):
    mocker.patch("ascmhl.__version__.ascmhl_tool_version", "0.0.1")

    requests_mock.get(
        "https://api.github.com/repos/ascmitc/mhl/releases/latest",
        exc=requests.exceptions.ConnectTimeout,
    )

    from importlib import reload
    import ascmhl.cli.update

    reload(ascmhl.cli.update)

    updater = ascmhl.cli.update.Updater()
    while not updater.finished:
        time.sleep(0.1)

    assert updater.latest_version is None
    assert updater.needs_update is False
    updater.join(timeout=1)



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from click.testing import CliRunner

import ascmhl.commands


def test_check_succeed(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["/root"])
    assert result.exit_code == 0


def test_check_fail_missing_history(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["/root"])
    assert result.exit_code == 11
    assert "/root" in result.output


def test_check_fail_altered_file(fs, simple_mhl_history):
    # alter a file
    with open("/root/Stuff.txt", "a") as file:
        file.write("!!")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["/root"])
    assert result.exit_code == 12
    assert "Stuff.txt" in result.output


def test_check_fail_new_file(fs, simple_mhl_history):
    # create a file not referenced in the history
    fs.create_file("/root/other.txt", contents="other\n")
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["/root"])
    assert result.exit_code == 13
    assert "other.txt" in result.output


def test_check_fail_missing_file(fs, simple_mhl_history):
    # remove a referenced file
    os.remove("/root/Stuff.txt")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["/root"])
    assert result.exit_code == 15
    assert "Stuff.txt" in result.output



#!/usr/bin/env python3
"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import click

from ascmhl import commands
from ascmhl.cli.update import Updater

updater = Updater()


class NaturalOrderGroup(click.Group):
    def list_commands(self, ctx):
        return self.commands.keys()


@click.group(cls=NaturalOrderGroup)
@click.version_option()
def mhltool_cli():
    pass


@mhltool_cli.resultcallback()
def update(*args, **kwargs):
    updater.join(timeout=1)
    if updater.needs_update:
        click.secho(f"Please update to the latest ascmhl version using `pip3 install -U ascmhl`.", fg="blue")


# new
mhltool_cli.add_command(commands.create)
mhltool_cli.add_command(commands.verify)
mhltool_cli.add_command(commands.diff)
mhltool_cli.add_command(commands.flatten)
mhltool_cli.add_command(commands.info)
mhltool_cli.add_command(commands.xsd_schema_check)


if __name__ == "__main__":
    mhltool_cli()



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2021, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from . import logger
from .__version__ import ascmhl_reference_hash_format
from .chain import MHLChain, MHLChainGeneration
from .hashlist import MHLHashList
import os
import textwrap

from lxml import etree
from lxml.builder import E

from .hashlist import *
from .__version__ import ascmhl_supported_hashformats
from .hashlist import MHLHashList


def parse(file_path):
    """parsing the MHL directory file and building the MHLChain for the chain member variable"""
    logger.debug(f'parsing "{os.path.basename(file_path)}"...')

    chain = MHLChain(file_path)
    chain.file_path = file_path
    current_object = None

    if not os.path.exists(file_path):
        return chain

    file = open(file_path, "rb")
    for event, element in etree.iterparse(file, events=("start", "end")):

        # check if we need to create a new container
        if event == "start":
            # the tag might contain the namespace like {urn:ASC:MHL:v2.0}hash, so we need to strip the namespace part
            # doing it with split is faster than using the lxml QName method
            tag = element.tag.split("}", 1)[-1]

            if not current_object:
                if tag == "hashlist":
                    current_object = MHLChainGeneration()

        elif event == "end":
            if current_object:
                tag = element.tag.split("}", 1)[-1]

                if type(current_object) is MHLChainGeneration:
                    if tag == "path":
                        current_object.ascmhl_filename = element.text
                    elif tag in ascmhl_supported_hashformats:
                        current_object.hash_format = tag
                        current_object.hash_string = element.text
                    elif tag == "hashlist":
                        current_object.generation_number = element.attrib.get("sequencer")
                        chain.append_generation(current_object)
                        current_object = None

    return chain


def write_chain(chain: MHLChain, new_hash_list: MHLHashList):
    logger.debug(f'writing "{os.path.basename(chain.file_path)}"...')

    """creates a new chain file and writes the xml to disk
    """

    directory_path = os.path.dirname(chain.file_path)
    if not os.path.isdir(directory_path):
        os.mkdir(directory_path)

    file = open(chain.file_path, "wb")
    file.write(b'<?xml version="1.0" encoding="UTF-8"?>\n<ascmhldirectory xmlns="urn:ASC:MHL:DIRECTORY:v2.0">\n')
    current_indent = "  "

    for generation in chain.generations:
        _write_xml_element_to_file(file, _hashlist_xml_element_from_chaingeneration(generation), "  ")

    # write new hashlist
    _write_xml_element_to_file(file, _hashlist_xml_element_from_hashlist(new_hash_list), "  ")

    current_indent = current_indent[:-2]
    _write_xml_string_to_file(file, "</ascmhldirectory>\n", current_indent)
    file.flush()


def _write_xml_element_to_file(file, xml_element, indent: str):
    xml_string = etree.tostring(xml_element, pretty_print=True, encoding="unicode")
    _write_xml_string_to_file(file, xml_string, indent)


def _write_xml_string_to_file(file, xml_string: str, indent: str):
    result = textwrap.indent(xml_string, indent)
    file.write(result.encode("utf-8"))


def _hashlist_xml_element_from_hashlist(hash_list: MHLHashList):
    """builds and returns one <hashlist> element for a given HashList object"""

    hash_list_element = E.hashlist(
        E.path(os.path.basename(hash_list.file_path)),
        E.c4(hash_list.generate_reference_hash()),
    )
    hash_list_element.attrib["sequencer"] = str(hash_list.generation_number)

    return hash_list_element


def _hashlist_xml_element_from_chaingeneration(generation: MHLChainGeneration):
    """builds and returns one <hashlist> element for a given ChainGeneration object"""

    if generation.hash_format == "c4":
        hash_list_element = E.hashlist(
            E.path(generation.ascmhl_filename),
            E.c4(generation.hash_string),
        )
        hash_list_element.attrib["sequencer"] = str(generation.generation_number)

        return hash_list_element
    else:
        logger.error("ERR: fixme: non-c4 hash in chain file, not implemented")
        return E.hashlist()
